"""jsonfile.py unit tests."""
from unittest.mock import patch

import pytest

import pypyr.parser.jsonfile


def test_json_file_open_fails_on_arbitrary_string():
    """Non path-y input string should fail."""
    with pytest.raises(FileNotFoundError):
        pypyr.parser.jsonfile.get_parsed_context('value 1,value 2, value3')


def test_json_file_open_fails_on_empty_string():
    """Non path-y input string should fail."""
    with pytest.raises(AssertionError):
        pypyr.parser.jsonfile.get_parsed_context(None)


def test_json_pass(fs):
    """Relative path to json should succeed."""
    in_path = './tests/testfiles/test.json'
    fs.create_file(in_path, contents="""{
    "key1": "value1",
    "key2": "value2",
    "key3": "value3"
}
""")
    context = pypyr.parser.jsonfile.get_parsed_context([in_path])

    assert context, "context shouldn't be None"
    assert len(context) == 3, "context should have 3 items"
    assert context["key2"] == "value2", "key2 should be value2"


@patch('pypyr.config.config.default_encoding', new='utf-16')
def test_json_pass_with_encoding(fs):
    """Relative path to json should succeed with encoding."""
    in_path = './tests/testfiles/test.json'
    fs.create_file(in_path, contents="""{
    "key1": "value1",
    "key2": "value2",
    "key3": "value3"
}
""", encoding='utf-16')

    context = pypyr.parser.jsonfile.get_parsed_context([in_path])

    assert context, "context shouldn't be None"
    assert len(context) == 3, "context should have 3 items"
    assert context["key2"] == "value2", "key2 should be value2"


def test_json_parse_not_mapping_at_root(fs):
    """Not mapping at root level raises."""
    in_path = './tests/testfiles/singleliteral.json'
    fs.create_file(in_path, contents='123')
    with pytest.raises(TypeError) as err_info:
        pypyr.parser.jsonfile.get_parsed_context([in_path])

    assert str(err_info.value) == (
        "json input should describe an object at the top "
        "level. You should have something like\n"
        "{\n\"key1\":\"value1\",\n\"key2\":\"value2\"\n}\n"
        "at the json top-level, not an [array] or literal.")



"""yamlfile.py unit tests."""
from unittest.mock import patch

import pytest

import pypyr.parser.yamlfile


def test_yaml_file_open_fails_on_arbitrary_string():
    """Non path-y input string should fail."""
    with pytest.raises(FileNotFoundError):
        pypyr.parser.yamlfile.get_parsed_context('value 1,value 2, value3')


def test_yaml_file_open_fails_on_empty_string():
    """Non path-y input string should fail."""
    with pytest.raises(AssertionError):
        pypyr.parser.yamlfile.get_parsed_context(None)


def test_yaml_pass(fs):
    """Relative path to yaml should succeed."""
    in_path = './tests/testfiles/dict.yaml'
    fs.create_file(in_path, contents="""key1: value1
key2: value2
key3: value3
key4:
  k41: k41value
  k42:
    - k42list1
    - k42list2
    - k42list3
  k43: True
  k44: 77
key5:
  - k5list1
  - k5list2
""")
    context = pypyr.parser.yamlfile.get_parsed_context([in_path])

    assert context, "context shouldn't be None"
    assert len(context) == 5, "context should have 5 items"
    assert context['key2'] == 'value2', "key2 should be value2"
    assert len(context['key4']['k42']) == 3, "3 items in k42"
    assert 'k42list2' in context['key4']['k42'], "k42 containts k42list2"
    assert context['key4']['k43'], "k43 is True"
    assert context['key4']['k44'] == 77, "k44 is 77"
    assert len(context['key5']) == 2, "2 items in key5"


@patch('pypyr.config.config.default_encoding', new='utf-16')
def test_yaml_pass_with_encoding(fs):
    """Relative path to yaml should succeed."""
    in_path = './tests/testfiles/dict.yaml'
    fs.create_file(in_path, contents="""key1: value1
key2: value2
key3: value3
key4:
  k41: k41value
  k42:
    - k42list1
    - k42list2
    - k42list3
  k43: True
  k44: 77
key5:
  - k5list1
  - k5list2
""", encoding='utf-16')

    context = pypyr.parser.yamlfile.get_parsed_context([in_path])

    assert context, "context shouldn't be None"
    assert len(context) == 5, "context should have 5 items"
    assert context['key2'] == 'value2', "key2 should be value2"
    assert len(context['key4']['k42']) == 3, "3 items in k42"
    assert 'k42list2' in context['key4']['k42'], "k42 containts k42list2"
    assert context['key4']['k43'], "k43 is True"
    assert context['key4']['k44'] == 77, "k44 is 77"
    assert len(context['key5']) == 2, "2 items in key5"


def test_list_yaml_fails(fs):
    """Yaml describing a list rather than a dict should fail."""
    in_path = './tests/testfiles/list.yaml'
    fs.create_file(in_path, contents="""- listitem1
- listitem2
  - listitem2.1
  - listitem2.2
- listitem3
""")
    with pytest.raises(TypeError):
        pypyr.parser.yamlfile.get_parsed_context([in_path])



"""Global cache for pypeloader modules.

Attributes:
    loader_cache: Global instance of the loader cache.
                  Use this attribute to access the cache from elsewhere.
"""
from collections.abc import Mapping
import logging

from pypyr.cache.cache import Cache
from pypyr.config import config
from pypyr.errors import PipelineDefinitionError
import pypyr.moduleloader
from pypyr.pipedef import PipelineDefinition, PipelineInfo

logger = logging.getLogger(__name__)


class Loader():
    """A single pipeline loader & the cache for all pipelines it has loaded.

    It loads pipelines using the get_pipeline_definition you assign to the
    loader at initialization.

    Attributes:
        name (str): Absolute module name of loader.
    """

    __slots__ = ['name', '_get_pipeline_definition', '_pipeline_cache']

    def __init__(self, name, get_pipeline_definition):
        """Initialize the loader and its pipeline cache.

        The expected function signature is:
        get_pipeline_definition(name: str,
                                parent: any) -> PipelineDefinition | Mapping

        Args:
            name: Absolute name of loader
            get_pipeline_definition: Reference to the function to call when
                loading a pipeline with this Loader.
        """
        self.name = name
        self._get_pipeline_definition = get_pipeline_definition
        self._pipeline_cache = Cache()

    def clear(self):
        """Clear all the pipelines in this Loader's cache."""
        self._pipeline_cache.clear()

    def get_pipeline(self, name, parent):
        """Get cached PipelineDefinition. Adds it to cache if it doesn't exist.

        The cache is local to this Loader instance.

        The combination of parent+name must be unique for this Loader. Parent
        should therefore have a sensible __str__ implementation because it
        forms part of the pipeline's identifying str key in the cache.

        Args:
            name (str): Name of pipeline, sans .yaml at end.
            parent (any): Parent in which to look for pipeline.

        Returns:
            pypyr.pipedef.PipelineDefinition: Yaml payload and loader info
                metadata for the pipeline.
        """
        # str keys perform better than tuples in dicts
        normalized_name = f'{parent}+{name}' if parent else name
        return self._pipeline_cache.get(
            normalized_name,
            lambda: self._load_pipeline(name, parent))

    def _load_pipeline(self, name, parent):
        """Execute get_pipeline_definition(name, parent) for this loader.

        If the loader get_pipeline_definition does not return a
        PipelineDefinition, this method will wrap the payload inside a
        PipelineDefinition for you.

        Args:
            name (str): Name of pipeline, sans .yaml at end.
            parent (any): Parent in which to look for pipeline.

        Returns:
            pypyr.pipedef.PipelineDefinition: Yaml payload and loader info
                metadata for the pipeline.
        """
        logger.debug("starting")

        logger.debug("loading the pipeline definition with %s", self.name)
        pipeline_definition = self._get_pipeline_definition(pipeline_name=name,
                                                            parent=parent)

        # loader can return either just the pipeline yaml payload itself, or a
        # PipelineDefinition wrapping both the pipeline yaml and its metadata.
        if not isinstance(pipeline_definition, PipelineDefinition):
            # standardize to the common cascading PipelineInfo if the loader
            # didn't provide any.
            pipeline_definition = PipelineDefinition(
                pipeline=pipeline_definition,
                info=PipelineInfo(pipeline_name=name,
                                  loader=self.name,
                                  parent=parent))

        # if the pipeline is not a dict at top level, failures down-stream get
        # mysterious - this prevents malformed pipe from even making it into
        # cache.
        if not isinstance(pipeline_definition.pipeline, Mapping):
            raise PipelineDefinitionError(
                "A pipeline must be a mapping at the top level. "
                "Does your top-level yaml have a 'steps:' key? "
                "For example:\n\n"
                "steps:\n"
                "  - name: pypyr.steps.echo\n"
                "    in:\n"
                "      echoMe: this is a bare bones pipeline example.\n")

        logger.debug("done")
        return pipeline_definition


class LoaderCache(Cache):
    """Get Loader instances from the pypeloader cache."""

    def get_pype_loader(self, loader=None):
        """Get cached Loader instance.

        Defaults to pypyr.loaders.file if loader not specified.

        Load the module specified by this absolute name, get its
        get_pipeline_definition function, wrap it in a Loader instance, and add
        it to cache.

        Args:
            loader (str): Absolute name of loader to use.

        Returns:
            pypyr.loadercache.Loader: Loader with a pointer to the
                get_pipeline_definition function found in the module specified
                by the loader parameter.
        """
        logger.debug("starting")

        if loader:
            logger.debug("you set the pype loader to: %s", loader)
        else:
            loader = config.default_loader
            logger.debug("use default pype loader: %s", loader)

        loader_instance = self.get(loader, lambda: load_the_loader(loader))

        logger.debug("done")
        return loader_instance

    def clear_pipes(self, loader_name=None):
        """Clear the pipeline cache.

        Args:
            loader_name (str): Clear pipelines for this loader. If not
                specified, will iterate all loaders in the loaders cache and
                clear each of their pipelines.
        """
        if loader_name:
            loader = self._cache.get(loader_name, None)
            if loader:
                loader.clear()
            else:
                logger.debug(
                    "%s not found in loader cache so there's nothing to clear",
                    loader_name)
        else:
            for _, loader in self._cache.items():
                loader.clear()


# global instance of the cache. use this to access the cache from elsewhere.
loader_cache = LoaderCache()


def load_the_loader(loader_name):
    """Load the module specified by loader and returns its function.

    Args:
        loader_name: string: name of module to load

    Returns:
        function: the get_pipeline_definition function in the module
    """
    logger.debug("starting")

    loader_module = pypyr.moduleloader.get_module(loader_name)

    try:
        get_pipeline_definition = getattr(loader_module,
                                          'get_pipeline_definition')
    except AttributeError:
        logger.error(
            "The pipeline loader %s doesn't have a "
            "get_pipeline_definition(pipeline_name, parent) "
            "function.",
            loader_module
        )
        raise

    loader = Loader(loader_name, get_pipeline_definition)
    logger.debug("%s done", loader_module)
    return loader



"""Control of flow instruction to switch between groups (if-else)."""
import logging
from pypyr.steps.dsl.cof import switch

logger = logging.getLogger(__name__)


def run_step(context):
    """Control of flow instruction to switch between step-groups.

    Call hands control back to this step upon completion, so execution
    continues from this point on-ward when the called groups are done.

    No fallthrough - will run the first matching group and not evaluate any
    case expressions thereafter.

    Args:
        context: pypyr.context.Context

    Context expects the following keys:
        switch:
            - case: expression. exec `call` when expression True.
              call: str. Single group-name to run.

            - default: exec `call` when none of the case expressions True.

    OR

        switch:
            - case: expression. exec call `when` expression True.
              call:
                groups: list[str] | str. Single group or list of groups to run.
                success: str. Name of group to run on success completion.
                failure: str. Name of group to run on something going wrong.

    `default` is optional. If you specify default, it must be the last item in
    the switch list.

    Return:
        None.
    """
    logger.debug("started")

    switch(context=context, name=__name__)

    # this will only run if no case condition true - else switch will kick out
    # of method with a raise.
    logger.debug("done")



"""Mock the pathlib Path library.

Uses delegation to preserve most Path functionality, e.g building, joining &
str representation of paths, but bypass actual filesystem operations.

Similar to "wraps" on the built-in Mock/MagicMock object in that you log all
calls to the object, but with more control over which Path method calls also to
pass-through to the underlying wrapped object.

Because of Path actually being a factory object that constructs the concrete
PosixPath/WindowsPath, the built-in Mock with wraps+side_effect overrides on
the return_value doesn't work too well - as soon as you try to set the
return_value of the resulting path object, Python ignores wraps
(as documented).
"""
from pathlib import Path
from unittest.mock import Mock


class FakePath():
    """Delegates the pathlib.Path object.

    Behaves like a Path, pass instance checks as Path, logs all calls like a
    Mock does. Overrides select methods NOT to touch the filesystem, while
    still passing logical operations like .join or .parent to the delegated
    Path object.

    fp = FakePath('/arb)
    # is_file handled by FakePath, not passed through to filesystem
    fp.is_file()

    Track method calls to this fake path like this:
    fp.mock.assert_called_once_with('/arb')

    This:
    fp.mock_instance.is_file.assert_called_once()

    Is equivalent of:
    fp.mock.return_value.is_file.assert_called_once()

    Attributes:
        path (Path): The pathlib Path instance being delegated
        mock (Mock): A Mock containing all the calls to the Path.
        mock_instance (Mock): Convenient access to mock.return_value.
    """

    def __init__(self, *pathsegments, is_dir=False, is_file=False):
        """Construct a fake Path.

        p = FakePath('/arb')

        Will create a delegated Path('p) object under the covers, while the
        delegating FakePath will behave like the actual Path object without
        interacting with the actual file-system.

        Args:
            pathsegments: Input to the underlying Path() object ctor. is_dir
            (bool): Faked result of is_dir(). is_file (bool): Faked result of
            is_file().
        """
        self.path = Path(*pathsegments)
        mock = Mock(spec=type(self.path))
        self.mock = mock
        self.mock_instance = mock(*pathsegments)
        self._is_dir = is_dir
        self._is_file = is_file

        self._attr_mock_cache = {}

    @property
    def __class__(self):
        """Pretend to be PosixPath or WindowsPath, depending on os."""
        return type(self.path)

    def __eq__(self, other):
        """Allow equality check to work as per path."""
        return self.path == other

    def __instancecheck__(self, instance):
        """Allow me to masquerade as an actual Path object."""
        return isinstance(instance, Path)

    def __fspath__(self):
        """Implement os.PathLike interface."""
        return str(self.path)

    def exists(self):
        """Fake exists not actually to touch the filesystem.

        Returns: (bool)
            Is True if either is_dir or is_file is True.
        """
        self.mock_instance.exists()
        return self._is_dir or self._is_file

    def is_dir(self):
        """Fake is_dir not actually to touch the filesystem."""
        self.mock_instance.is_dir()
        return self._is_dir

    def is_file(self):
        """Fake is_file not actually to touch the filesystem."""
        self.mock_instance.is_file()
        return self._is_file

    def mkdir(self, *args, **kwargs):
        """Fake mkdir not actually to touch the filesystem."""
        self.mock_instance.mkdir(*args, **kwargs)

    def __getattr__(self, attr):
        """Delegate not found attrs to underlying path instance."""
        # Both mock the method call AND pass through to delegated Path method.
        # The attr name is unique for this object, so you can use the same mock
        # to log all calls to the same method.
        m = self._attr_mock_cache.get(attr, None)
        if m is None:
            m = Mock()()
            self._attr_mock_cache[attr] = m
            self.mock_instance.attach_mock(m, attr)

        underlying_attr = getattr(self.path, attr)

        if not callable(underlying_attr):
            return FakePath(underlying_attr) if isinstance(
                underlying_attr, Path) else underlying_attr

        # wrap the delegated callable so you can
        # capture the args+kwargs when callable
        # actually called
        def passthrough_wrapper(*args, **kwargs):
            m(*args, **kwargs)
            return underlying_attr(*args, **kwargs)

        return passthrough_wrapper

    def __str__(self):
        """Delegate str to underlying path object."""
        return str(self.path)


class MultiPath():
    """Use me as a `new` replacement on patch for multiple Path instantiations.

    This class will masquerade as a Path instance so this is True:
    isinstance(MultiPath(), Path)

    Use me like this:
        with patch('mymoduleundertest.Path', new=MultiPath()) as mock_path:
            do_things()

        # all the calls to the mock here, for your asserts etc.
        mock_path.instances[0].mock.mock_calls

    mock_path.instances contains all the mocked Path objects created in
            do_things() in order created.

    NB: use with "new" on patch, NOT with "side_effect".

    Attributes:
        instances (list): Instances of FakePath constructed under the patch
            scope, in order.
        known (dict): Cache of preconfigured FakePath to return when the
            input arguments match.
    """

    def __init__(self, known=None):
        """Construct the callable object to replace the patched Path.

        Use known to pre-configure FakePath instances for certain input
        arguments if you want to override FakePath defaults. The key of known
        is the pathsegments input of the Path call you want to mock.

        For example:
        {'p': FakePath(is_file=True), ('p1, 'p2): FakePath(is_dir=True)}

        This will use pre-created cached instances for:
            - Path('p')
            - Path('p1', 'p2)

        If you don't use known, will just create a FakePath with defaults.

        Args:
            known (dict): Cache of known FakePath instances mapped to their
                input args.
        """
        self.instances = []
        self.known = known if known else {}

    def __instancecheck__(self, instance):
        """Trick me into passing Path isinstance checks."""
        return isinstance(instance, FakePath)

    @property
    def __class__(self):
        """Trick me into looking like a Path."""
        return FakePath

    def __call__(self, *p):
        """Call this everytime the patched object instantiates.

        Will return either a new default FakePath or a cached FakePath from
        self.known if *p matches on each invocation.

        Args:
            *p: pathsegments input for mocked Path object.

        Returns:
            FakePath instance.
        """
        # this allows cache to hit on Path('one', 'two') and Path('one')
        fp = self.known.get(p[0]) if len(p) == 1 else self.known.get(p)
        if not fp:
            fp = FakePath(*p)

        self.instances.append(fp)
        return fp



"""pypyr step that executes commands in the shell as a sub-process.

The command string must be formatted exactly as it would be when typed
at the shell prompt. This includes, for example, quoting or backslash escaping
filenames with spaces in them. The shell defaults to /bin/sh on posix.
"""
import logging
from pypyr.steps.dsl.cmdasync import AsyncCmdStep

logger = logging.getLogger(__name__)


def run_step(context):
    """Run commands, programs or executables asynchronously, in parallel.

    Spawns a shell for each runnable item.

    Context is a pypyr.context.Context. This is dict-like.

    Context must contain the following keys:
        cmds:
            - <<shell string 1>>
            - <<shell string 2>>

    where <<cmd string>> is command + args to execute.

    You can alternatively use the expanded syntax to override default options:

    cmds:
        run: list[str | list[str]]. mandatory. command + args to execute.
            If list entry is another list[str], the sub-list will run in
            serial.
        save: bool. defaults False. save output to cmdOut. Treats output
            as text in the system's encoding and removes newlines at end.
        cwd: str/Pathlike. optional. Working directory for these commands.
        bytes (bool): Default False. When `save` return output bytes from
            cmds unaltered, without applying any encoding & text newline
            processing.
        encoding (str): Default None. When `save`, decode output with
            this encoding. The default of None uses the system encoding and
            should "just work".
        stdout (str | Path): Default None. Write stdout to this file path.
            Special value `/dev/null` writes to the system null device.
        stderr (str | Path): Default None. Write stderr to this file path.
            Special value `/dev/null` writes to the system null device.
            Special value `/dev/stdout` redirects err output to stdout.
        append (bool): Default False. When stdout/stderr a file, append
            rather than overwrite. Default is to overwrite.

    In expanded syntax, `run` can be a simple string or a list:
        cmds:
          run:
            - ./my-executable --arg
            - ./another-executable --arg
          save: False
          cwd: ./path/here

    Will execute the command string in the shell as a sub-process.
    Escape curly braces: if you want a literal curly brace, double it like
    {{ or }}.

    If save is True, will save the output as a list of results in the order
    executed. Each result is a pypyr.subproc.SubprocessResult, with the schema:
    cmd: the cmd/args executed
    returncode: 0
    stdout: 'stdout str here. None if empty.'
    stderr: 'stderr str here. None if empty.'

    cmdOut.returncode is the exit status of the called process. Typically 0
    means OK. A negative value -N indicates that the child was terminated by
    signal N (POSIX only).

    If the sub-process couldn't spawn at all (e.g executable not found), the
    results list will contain the Exception object instead of a
    SubprocessResult.

    You can find this list of results in context `cmdOut` after step completes.

    context['cmds'] will interpolate anything in curly braces for values
    found in context. So if your context looks like this:
        key1: value1
        key2: value2
        cmds:
            - mything --arg1 {key1}

    The cmd passed to the shell will be "mything --arg value1"
    """
    logger.debug("started")

    AsyncCmdStep(name=__name__, context=context, is_shell=True).run_step()

    logger.debug("done")



"""pypyr.dsl.cmdasync unit tests."""
# can remove __future__ once py 3.10 the lowest supported version
from __future__ import annotations
import asyncio
import asyncio.subprocess
import locale
from pathlib import Path
from unittest import TestCase
from unittest.mock import call, DEFAULT, Mock, mock_open, patch

import pytest

from pypyr.config import config
from pypyr.context import Context
from pypyr.errors import (ContextError,
                          KeyInContextHasNoValueError,
                          KeyNotInContextError,
                          MultiError,
                          SubprocessError)
from pypyr.steps.dsl.cmdasync import AsyncCmdStep
from pypyr.aio.subproc import Command

PIPE = asyncio.subprocess.PIPE
LOCALE_ENCODING = locale.getpreferredencoding(False)

cmd_path = Path.cwd().joinpath('tests/testfiles/cmds')
is_windows = config.is_windows
is_posixy = not is_windows


def get_cmd(posix, win):
    """Return posix or win depending on current platform."""
    return win if is_windows else posix

# region validation errors


def test_dsl_async_cmd_name_required():
    """Cmd Step requires name."""
    with pytest.raises(AssertionError):
        AsyncCmdStep(None, None)


def test_dsl_async_cmd_context_required():
    """Cmd Step requires context."""
    with pytest.raises(AssertionError):
        AsyncCmdStep('blah', None)


def test_dsl_async_cmd_context_cmd_required():
    """Cmd Step requires cmds in context."""
    with pytest.raises(KeyNotInContextError) as err:
        AsyncCmdStep('blah', Context({'a': 'b'}))

    assert str(err.value) == ("context['cmds'] doesn't exist. It must exist "
                              "for blah.")


def test_dsl_async_cmd_context_cmd_not_none():
    """Cmd Step requires cmds in context."""
    with pytest.raises(KeyInContextHasNoValueError) as err:
        AsyncCmdStep('blah', Context({'cmds': None}))

    assert str(err.value) == "context['cmds'] must have a value for blah."


def test_dsl_async_cmd_context_cmd_not_dict():
    """Cmd Step requires cmds in context to be a dict if not str."""
    with pytest.raises(ContextError) as err:
        AsyncCmdStep('blah', Context({'cmds': 1}))

    assert str(err.value) == (
        """\
blah cmds config should be either a list:
cmds:
  - ./my-executable --arg
  - subdir/executable --arg1

or a dictionary with a `run` sub-key:
cmds:
  run:
    - ./my-executable --arg
    - subdir/executable --arg1
  cwd: ./mydir

Any of the list items in root can be in expanded syntax:
cmds:
  - ./my-executable --arg
  - subdir/executable --arg1
  - run:
      - ./arb-executable1 --arg value1
      - [./arb-executable2.1, ./arb-executable2.2]
    cwd: ../mydir/subdir
  - [./arb-executable3.1, ./arb-executable3.2]""")


def test_dsl_async_cmd_list_must_be_str_or_dict():
    """Each list input must be a string or a dict."""
    with pytest.raises(ContextError) as err:
        AsyncCmdStep('blah', Context({'cmds': ['cmd1', 123]}))

    assert str(err.value) == ("""\
123 in blah cmds config is wrong.
Each list item should be either a simple string, or a list to run in serial,
or a dict for expanded syntax:
cmds:
  - ./my-executable --arg
  - run:
      - ./another-executable --arg value
      - ./another-executable --arg value2
    cwd: ../mydir/subdir
  - run:
      - ./arb-executable1 --arg value1
      - [./arb-executable2.1, ./arb-executable2.2]
    cwd: ../mydir/arbdir
  - [./arb-executable3.1, ./arb-executable3.2]""")


def test_dsl_async_cmd_dict_run_must_exist():
    """Dict input run must exist."""
    with pytest.raises(ContextError) as err:
        AsyncCmdStep('blah', Context({'cmds': {'runs': 'abc'}}))

    # noqa is for line too long
    assert str(err.value) == ("""\
cmds.run doesn't exist for blah.
The input should look like this in expanded syntax:
cmds:
  run:
    - ./my-executable --arg
    - subdir/executable --arg1
  cwd: ./mydir

If you're passing in a list of commands, each command should be a simple string,
or a sub-list of commands to run in serial,
or a dict with a `run` entry:
cmds:
  - ./my-executable --arg
  - run: ./another-executable --arg value
    cwd: ../mydir/subdir
  - run:
      - ./arb-executable1 --arg value1
      - [./arb-executable2.1, ./arb-executable2.2]
    cwd: ../mydir/arbdir
  - [./arb-executable3.1, ./arb-executable3.2]""")  # noqa: E501


def test_dsl_async_cmd_dict_run_must_have_value():
    """Dict input run must have value."""
    with pytest.raises(ContextError) as err:
        AsyncCmdStep('blah', Context({'cmds': {'run': ''}}))

    assert str(err.value) == ("""\
cmds.run must have a value for blah.
The `run` input should look something like this:
cmds:
  run:
    - ./arb-executable1 --arg value1
    - ./arb-executable2 --arg value2
  cwd: ../mydir/arbdir""")


def test_dsl_async_cmd_must_be_str_or_list():
    """Input to cmd must be a str or a list."""
    with pytest.raises(MultiError) as err:
        cmd = AsyncCmdStep('blah', Context({'cmds': {'run': 123}}))
        cmd.run_step()

    assert str(err.value) == ("""\
The following error(s) occurred while running the async commands:
ContextError: 123 cmds input.
Each command should be a simple string, or a sub-list to run in serial:
cmds:
  - ./my-executable --arg
  - ./another-executable --arg
  - [./executableA.1, ./executableA.2]

Or in the expanded syntax, set `run` to a list:
cmds:
  run:
    - ./my-executable --arg
    - ./another-executable --arg
    - [./executableA.1, ./executableA.2]
  cwd: ./mydir""")


def test_dsl_async_cmd_run_save_with_stdout_stderr():
    """Raise err when setting stdout | stderr alongside save."""
    context = Context({
        'cmds': {
            'run': ['A', 'B'],
            'save': True,
            'stdout': '/arb1',
            'stderr': '/arb2'}
    })

    with pytest.raises(ContextError) as err:
        AsyncCmdStep('blah', context)

    assert str(err.value) == (
        "You can't set `stdout` or `stderr` when `save` is True.")


def test_dsl_async_cmd_run_save_with_stdout():
    """Raise err when setting stdout | stderr alongside save."""
    context = Context({
        'cmds': {
            'run': ['A', 'B'],
            'save': True,
            'stdout': '/arb1'}
    })

    with pytest.raises(ContextError) as err:
        AsyncCmdStep('blah', context)

    assert str(err.value) == (
        "You can't set `stdout` or `stderr` when `save` is True.")


def test_dsl_async_run_save_with_stdout_or_stderr():
    """Raise err when setting stdout | stderr alongside save - or not and."""
    context = Context({
        'cmds': {
            'run': ['A', 'B'],
            'save': True,
            'stderr': '/arb2'}
    })

    with pytest.raises(ContextError) as err:
        AsyncCmdStep('blah', context)

    assert str(err.value) == (
        "You can't set `stdout` or `stderr` when `save` is True.")

# endregion validation errors

# region minimal/maximal inputs


def test_async_cmd_minimal():
    """Minimal inputs."""
    context = Context({'cmds': ['A', 'B']})

    step = AsyncCmdStep('blah', context)

    assert len(step.commands) == 2
    cmd1 = step.commands[0]
    assert cmd1.cmd == 'A'
    assert cmd1.is_shell is False
    assert cmd1.cwd is None
    assert cmd1.is_save is False
    assert cmd1.is_text is False
    assert cmd1.stdout is None
    assert cmd1.stderr is None
    assert cmd1.encoding == LOCALE_ENCODING
    assert cmd1.append is False


def test_async_cmd_minimal_save():
    """Minimal inputs when save True."""
    context = Context({'cmds': {'run': ['A', 'B'], 'save': True}})

    step = AsyncCmdStep('blah', context)

    assert len(step.commands) == 1
    cmd1 = step.commands[0]
    assert cmd1.cmd == ['A', 'B']
    assert cmd1.is_shell is False
    assert cmd1.cwd is None
    assert cmd1.is_save is True
    assert cmd1.is_text is True
    assert cmd1.stdout == PIPE
    assert cmd1.stderr == PIPE
    assert cmd1.encoding == LOCALE_ENCODING
    assert cmd1.append is False


def test_async_cmd_maximal_save():
    """Maximal inputs with save true."""
    context = Context({'cmds': {
        'run': ['A', 'B'],
        'cwd': '/cwd',
        'save': True,
        'encoding': 'enc',
        'bytes': True}})

    step = AsyncCmdStep('blah', context)

    assert len(step.commands) == 1
    cmd1 = step.commands[0]
    assert cmd1.cmd == ['A', 'B']
    assert cmd1.is_shell is False
    assert cmd1.is_text is False
    assert cmd1.cwd == '/cwd'
    assert cmd1.is_save is True
    assert cmd1.stdout == PIPE
    assert cmd1.stderr == PIPE
    assert cmd1.encoding == 'enc'
    assert cmd1.append is False


def test_async_cmd_maximal_not_save():
    """Maximal inputs with save false."""
    context = Context({'cmds': {
        'run': ['A', 'B'],
        'cwd': '/cwd',
        'stdout': '/stdout',
        'stderr': '/stderr',
        'encoding': 'enc',
        'bytes': True,
        'append': True}})

    step = AsyncCmdStep('blah', context)

    assert len(step.commands) == 1
    cmd1 = step.commands[0]
    assert cmd1.cmd == ['A', 'B']
    assert cmd1.is_shell is False
    assert cmd1.is_text is False
    assert cmd1.cwd == '/cwd'
    assert cmd1.is_save is False
    assert cmd1.stdout == '/stdout'
    assert cmd1.stderr == '/stderr'
    assert cmd1.encoding == 'enc'
    assert cmd1.append is True


def test_async_cmd_maximal_save_formatting():
    """Maximal inputs with save true with formatting expressions."""
    context = Context({
        'cmdA': 'A',
        'cmdB': 'B',
        'cwd': '/cwd',
        'save': True,
        'enc': 'enc',
        'bytes': True,
        'cmds': {
            'run': ['{cmdA}', '{cmdB}'],
            'cwd': '{cwd}',
            'save': '{save}',
            'encoding': '{enc}',
            'bytes': '{bytes}'}})

    step = AsyncCmdStep('blah', context)

    assert len(step.commands) == 1
    cmd1 = step.commands[0]
    assert cmd1.cmd == ['A', 'B']
    assert cmd1.is_shell is False
    assert cmd1.is_text is False
    assert cmd1.cwd == '/cwd'
    assert cmd1.is_save is True
    assert cmd1.stdout == PIPE
    assert cmd1.stderr == PIPE
    assert cmd1.encoding == 'enc'
    assert cmd1.append is False


def test_async_cmd_maximal_not_save_formatting():
    """Maximal inputs with save false with formatting expressions."""
    context = Context({
        'cmdinput': ['A', 'B'],
        'cwd': '/cwd',
        'save': True,
        'enc': 'enc',
        'bytes': True,
        'stdout': '/stdout',
        'stderr': '/stderr',
        'append': True,
        'cmds': {
            'run': '{cmdinput}',
            'cwd': '{cwd}',
            'stdout': '{stdout}',
            'stderr': '{stderr}',
            'encoding': '{enc}',
            'bytes': '{bytes}',
            'append': '{append}'}})

    step = AsyncCmdStep('blah', context)

    assert len(step.commands) == 1
    cmd1 = step.commands[0]
    assert cmd1.cmd == ['A', 'B']
    assert cmd1.is_shell is False
    assert cmd1.is_text is False
    assert cmd1.cwd == '/cwd'
    assert cmd1.is_save is False
    assert cmd1.stdout == '/stdout'
    assert cmd1.stderr == '/stderr'
    assert cmd1.encoding == 'enc'
    assert cmd1.append is True

# endregion minimal/maximal inputs

# region test input combos w mocked subprocess call

# region AsyncMock plumbing


class AsyncMock(Mock):
    """Mock that acts like a coroutine/awaitable/future.

    As of py3.8, could use the built-in AsyncMock from unittest. But, still
    have to cater to py3.7, so stuck with this, yay.
    """

    async def __call__(self, *args, **kwargs):
        """Make the mock awaitable."""
        return super().__call__(*args, **kwargs)


def get_async_subproc_future(expected: list[asyncio.subprocess.Process]):
    """Return awaitable mock with canned responses/results."""
    return AsyncMock(side_effect=expected)


def expected_result(rc, stdout, stderr):
    """Return a mock of Process with preconfigured output for communicate()."""
    process = Mock(spec=asyncio.subprocess.Process)
    process.returncode = rc
    process.communicate = AsyncMock(return_value=(stdout, stderr))
    return process


ASYNC_SUBPROCESS_EXEC = 'pypyr.aio.subproc.asyncio.create_subprocess_exec'
ASYNC_SUBPROCESS_SHELL = 'pypyr.aio.subproc.asyncio.create_subprocess_shell'

# endregion AsyncMock plumbing


def test_async_cmd_simple_list():
    """Simple syntax list pass through with split vs no split on shlex."""
    context = Context({'cmds': ['A', 'B arg1 "arg 2"', 'C']})
    step = AsyncCmdStep('blah', context)

    assert step.commands.commands == [Command('A'),
                                      Command('B arg1 "arg 2"'),
                                      Command('C')]

    fake_subproc = get_async_subproc_future(
        [expected_result(0, None, None)] * 3)

    with patch(ASYNC_SUBPROCESS_EXEC, fake_subproc) as mock_subproc:
        step.run_step()

    # despite name, checks equality for all items in ANY order
    TestCase().assertCountEqual(mock_subproc.mock_calls, [
        call('A', stdout=None, stderr=None, cwd=None),
        call('B', 'arg1', 'arg 2', stdout=None, stderr=None, cwd=None),
        call('C', stdout=None, stderr=None, cwd=None)])

    assert 'cmdOut' not in context


def test_async_cmd_simple_str():
    """Simple syntax str for single item."""
    context = Context({'cmds': 'A'})
    step = AsyncCmdStep('blah', context)

    assert step.commands.commands == [Command('A')]

    fake_subproc = get_async_subproc_future([expected_result(0, None, None)])

    with patch(ASYNC_SUBPROCESS_EXEC, fake_subproc) as mock_subproc:
        step.run_step()

    assert mock_subproc.mock_calls == [
        call('A', stdout=None, stderr=None, cwd=None)]

    assert 'cmdOut' not in context


def test_async_shell_simple_list():
    """Simple syntax list pass through with split vs no split on shlex."""
    context = Context({'cmds': ['A', 'B arg1 "arg 2"', 'C']})
    step = AsyncCmdStep('blah', context, is_shell=True)

    assert step.commands.commands == [Command('A', is_shell=True),
                                      Command('B arg1 "arg 2"', is_shell=True),
                                      Command('C', is_shell=True)]

    fake_subproc = get_async_subproc_future(
        [expected_result(0, None, None)] * 3)

    with patch(ASYNC_SUBPROCESS_SHELL, fake_subproc) as mock_subproc:
        step.run_step()

    # despite name, checks equality for all items in ANY order
    TestCase().assertCountEqual(mock_subproc.mock_calls, [
        call('A', stdout=None, stderr=None, cwd=None),
        call('B arg1 "arg 2"', stdout=None, stderr=None, cwd=None),
        call('C', stdout=None, stderr=None, cwd=None)])

    assert 'cmdOut' not in context


def test_async_list_with_dict():
    """Simple syntax list with any item a dict."""
    context = Context({'cmds': [
        'A',
        {'run': ['B', 'C']},
        {'run': ['D', ['E.1', 'E.2'], 'F']},
        'G']})
    step = AsyncCmdStep('blah', context)

    assert step.commands.commands == [Command('A'),
                                      Command(['B', 'C']),
                                      Command(['D', ['E.1', 'E.2'], 'F']),
                                      Command('G')]

    fake_subproc = get_async_subproc_future(
        [expected_result(0, None, None)] * 8)

    with patch(ASYNC_SUBPROCESS_EXEC, fake_subproc) as mock_subproc:
        step.run_step()

    # despite name, checks equality for all items in ANY order
    # sadly can't really confirm that the serial calls happened serially here
    TestCase().assertCountEqual(mock_subproc.mock_calls, [
        call('A', stdout=None, stderr=None, cwd=None),
        call('B', stdout=None, stderr=None, cwd=None),
        call('C', stdout=None, stderr=None, cwd=None),
        call('D', stdout=None, stderr=None, cwd=None),
        call('E.1', stdout=None, stderr=None, cwd=None),
        call('E.2', stdout=None, stderr=None, cwd=None),
        call('F', stdout=None, stderr=None, cwd=None),
        call('G', stdout=None, stderr=None, cwd=None)])

    assert 'cmdOut' not in context


def test_async_simple_list_with_serial():
    """Simple syntax list with sub-list for serial."""
    context = Context({'cmds': ['A', ['B.1', 'B.2'], ['C.1', 'C.2'], 'D']})
    step = AsyncCmdStep('blah', context)

    assert step.commands.commands == [Command('A'),
                                      Command([['B.1', 'B.2']]),
                                      Command([['C.1', 'C.2']]),
                                      Command('D')]

    fake_subproc = get_async_subproc_future(
        [expected_result(0, None, None)] * 6)

    with patch(ASYNC_SUBPROCESS_EXEC, fake_subproc) as mock_subproc:
        step.run_step()

    # despite name, checks equality for all items in ANY order
    # sadly can't really confirm that the serial calls happened serially here
    TestCase().assertCountEqual(mock_subproc.mock_calls, [
        call('A', stdout=None, stderr=None, cwd=None),
        call('D', stdout=None, stderr=None, cwd=None),
        call('B.1', stdout=None, stderr=None, cwd=None),
        call('B.2', stdout=None, stderr=None, cwd=None),
        call('C.1', stdout=None, stderr=None, cwd=None),
        call('C.2', stdout=None, stderr=None, cwd=None)])

    assert 'cmdOut' not in context


def test_async_dict_run_list_with_serial():
    """Expanded syntax run list with sub-list for serial."""
    context = Context({'cmds': {'run':
                                ['A', ['B.1', 'B.2'], ['C.1', 'C.2'], 'D']}})
    step = AsyncCmdStep('blah', context)

    assert step.commands.commands == [Command(['A',
                                      ['B.1', 'B.2'],
                                      ['C.1', 'C.2'],
                                      'D'
                                      ])]

    fake_subproc = get_async_subproc_future(
        [expected_result(0, None, None)] * 6)

    with patch(ASYNC_SUBPROCESS_EXEC, fake_subproc) as mock_subproc:
        step.run_step()

    # despite name, checks equality for all items in ANY order
    # sadly can't really confirm that the serial calls happened serially here
    TestCase().assertCountEqual(mock_subproc.mock_calls, [
        call('A', stdout=None, stderr=None, cwd=None),
        call('D', stdout=None, stderr=None, cwd=None),
        call('B.1', stdout=None, stderr=None, cwd=None),
        call('B.2', stdout=None, stderr=None, cwd=None),
        call('C.1', stdout=None, stderr=None, cwd=None),
        call('C.2', stdout=None, stderr=None, cwd=None)])

    assert 'cmdOut' not in context

# region save


def test_dsl_async_cmd_shell_override():
    """Override shell arg from dict shell input."""
    obj = AsyncCmdStep('blahname', Context({'cmds': {'run': 'blah',
                                                     'cwd': 'pathhere',
                                                     'shell': True,
                                                     }}),
                       is_shell=False)

    assert not obj.is_shell
    assert obj.logger.name == 'blahname'
    assert len(obj.commands.commands) == 1
    cmd = obj.commands.commands[0]
    assert cmd.is_shell
    assert not cmd.is_save


def test_async_cmd_with_save():
    """Save with simple list."""
    context = Context({'cmds': {'run': ['A', '"B with space" --arg', 'C'],
                                'save': True}})
    step = AsyncCmdStep('blah', context)

    assert step.commands.commands == [Command(['A',
                                      '"B with space" --arg',
                                               'C'], is_save=True)]

    fake_subproc = get_async_subproc_future([
        expected_result(0, b'out1', b'err1'),
        expected_result(0, b'out2', b'err2'),
        expected_result(0, b'out3', b'err3')])

    with patch(ASYNC_SUBPROCESS_EXEC, fake_subproc) as mock_subproc:
        step.run_step()

    # despite name, checks equality for all items in ANY order
    TestCase().assertCountEqual(mock_subproc.mock_calls, [
        call('A', stdout=PIPE, stderr=PIPE, cwd=None),
        call('B with space', '--arg', stdout=PIPE, stderr=PIPE, cwd=None),
        call('C', stdout=PIPE, stderr=PIPE, cwd=None)])

    out1 = context['cmdOut'][0]
    assert out1.cmd == ['A']
    assert out1.returncode == 0
    assert out1.stdout == 'out1'
    assert out1.stderr == 'err1'

    out2 = context['cmdOut'][1]
    assert out2.cmd == ['B with space', '--arg']
    assert out2.returncode == 0
    assert out2.stdout == 'out2'
    assert out2.stderr == 'err2'

    out3 = context['cmdOut'][2]
    assert out3.cmd == ['C']
    assert out3.returncode == 0
    assert out3.stdout == 'out3'
    assert out3.stderr == 'err3'


def test_async_shell_with_save():
    """Save with simple list."""
    context = Context({'cmds': {'run': ['A', '"B with space" --arg', 'C'],
                                'save': True}})
    step = AsyncCmdStep('blah', context, is_shell=True)

    assert step.commands.commands == [Command(['A',
                                      '"B with space" --arg',
                                               'C'],
                                              is_shell=True,
                                              is_save=True)]

    fake_subproc = get_async_subproc_future([
        expected_result(0, b'out1', b'err1'),
        expected_result(0, b'out2', b'err2'),
        expected_result(0, b'out3', b'err3')])

    with patch(ASYNC_SUBPROCESS_SHELL, fake_subproc) as mock_subproc:
        step.run_step()

    # despite name, checks equality for all items in ANY order
    TestCase().assertCountEqual(mock_subproc.mock_calls, [
        call('A', stdout=PIPE, stderr=PIPE, cwd=None),
        call('"B with space" --arg', stdout=PIPE, stderr=PIPE, cwd=None),
        call('C', stdout=PIPE, stderr=PIPE, cwd=None)])

    out1 = context['cmdOut'][0]
    assert out1.cmd == 'A'
    assert out1.returncode == 0
    assert out1.stdout == 'out1'
    assert out1.stderr == 'err1'

    out2 = context['cmdOut'][1]
    assert out2.cmd == '"B with space" --arg'
    assert out2.returncode == 0
    assert out2.stdout == 'out2'
    assert out2.stderr == 'err2'

    out3 = context['cmdOut'][2]
    assert out3.cmd == 'C'
    assert out3.returncode == 0
    assert out3.stdout == 'out3'
    assert out3.stderr == 'err3'


# endregion save

def test_async_cmds_fail_to_open_file_handle():
    """Catch error when file handle can't open."""
    context = Context({
        'cmds': {
            'run': ['cmd1', 'cmd2'],
            'stdout': '/stdout',
            'stderr': '/stderr'}
    })

    step = AsyncCmdStep('blah', context)

    with patch('pypyr.aio.subproc.open', mock_open()) as mock_fs:
        mock_fs.side_effect = [DEFAULT, FileNotFoundError('two')]

        with pytest.raises(MultiError) as err:
            with patch(ASYNC_SUBPROCESS_EXEC) as mock_subproc:
                step.run_step()

            mock_subproc.assert_not_called()

        the_err = err.value
        assert len(the_err) == 1
        assert str(the_err[0]) == 'two'

    results = step.commands[0].results
    assert len(results) == 1
    assert repr(results[0]) == "FileNotFoundError('two')"


def test_async_cmds_fail_to_open_file_handle_multiple_files():
    """Catch error when file handle can't open on multiple expanded."""
    context = Context({
        'cmds': [{
            'run': ['cmd1', 'cmd2'],
            'stdout': '/stdout',
            'stderr': '/stderr'},
            {
            'run': ['cmd3', 'cmd4'],
            'stdout': '/stdout2',
            'stderr': '/stderr2'}
        ]})

    step = AsyncCmdStep('blah', context)

    with patch('pypyr.aio.subproc.open', mock_open()) as mock_fs:
        mock_fs.side_effect = [DEFAULT,
                               FileNotFoundError('two'),
                               FileNotFoundError('three'),
                               DEFAULT]

        with pytest.raises(MultiError) as err:
            with patch(ASYNC_SUBPROCESS_EXEC) as mock_subproc:
                step.run_step()

            mock_subproc.assert_not_called()

        the_err = err.value
        assert len(the_err) == 2
        assert str(the_err[0]) == 'two'
        assert str(the_err[1]) == 'three'

    assert len(step.commands) == 2
    results = step.commands[0].results
    assert len(results) == 1
    assert repr(results[0]) == "FileNotFoundError('two')"

    results = step.commands[1].results
    assert len(results) == 1
    assert repr(results[0]) == "FileNotFoundError('three')"


def test_async_cmd_serial_simple_sequence():
    """Serial happens one after the other."""
    context = Context({'cmds': ['A', ['B.1', 'B.2'], 'C']})

    step = AsyncCmdStep('blah', context)

    fake_subproc = get_async_subproc_future([
        expected_result(0, b'outA', b'errA'),
        expected_result(0, b'outB.1', b'errB.1'),
        expected_result(0, b'outC', b'errC'),
        expected_result(0, b'outB.2', b'errB.2'), ])

    with patch(ASYNC_SUBPROCESS_EXEC, fake_subproc) as mock_subproc:
        step.run_step()

    assert len(mock_subproc.mock_calls) == 4
    TestCase().assertCountEqual(mock_subproc.mock_calls[:3], [
        call('A', stdout=None, stderr=None, cwd=None),
        call('B.1', stdout=None, stderr=None, cwd=None),
        call('C', stdout=None, stderr=None, cwd=None)
    ])

    assert mock_subproc.mock_calls[3] == call('B.2',
                                              stdout=None, stderr=None,
                                              cwd=None)
    assert 'cmdOut' not in context
# endregion test input combos w mocked subprocess call

# region actual subprocess spawning


def test_dsl_async_cmd_str_with_spaces_in_path():
    """Single command string works with spaces in path."""
    cmd = get_cmd('tests/testfiles/cmds/"file with space.sh"',
                  r'tests\testfiles\cmds\file with space.bat')
    context = Context({'a': 'one two',
                       'd': cmd,
                      'cmds': '{d} "{a}"'})

    step = AsyncCmdStep('blah', context)
    step.run_step()

    assert 'cmdOut' not in context

    # and with quotes
    if is_windows:
        context = Context({
            'cmds': r'"tests\testfiles\cmds\file with space.bat" "one two"'})
        step = AsyncCmdStep('blah', context)
        step.run_step()

        assert 'cmdOut' not in context


def test_dsl_async_cmd_error_throws():
    """Process returning 1 should throw CalledProcessError with no save."""
    cmd1 = get_cmd('tests/testfiles/cmds/exit1.sh',
                   r'tests\testfiles\cmds\exit1.bat')

    cmd2 = get_cmd('tests/testfiles/cmds/exitwitherr.sh',
                   r'tests\testfiles\cmds\exitwitherr.bat')

    context = Context({'cmds': [cmd1, cmd2]})
    step = AsyncCmdStep('blah', context)

    with pytest.raises(MultiError) as multi_err:
        step.run_step()

    assert 'cmdOut' not in context

    errs = multi_err.value
    assert len(errs) == 2

    err1 = errs[0]
    assert type(err1) is SubprocessError
    assert err1.cmd == [cmd1]

    err2 = errs[1]
    assert type(err2) is SubprocessError
    assert err2.cmd == [cmd2]


def test_dsl_async_cmd_error_throws_exception_initiating_spawn():
    """Exception when can't run the subprocess at all."""
    cmd1 = get_cmd('tests/testfiles/cmds/xxx1',
                   r'tests\testfiles\cmds\xxx1')

    cmd2 = get_cmd('tests/testfiles/cmds/xxx2',
                   r'tests\testfiles\cmds\xxx2')

    cmd3 = get_cmd('tests/testfiles/cmds/exit1.sh',
                   r'tests\testfiles\cmds\exit1.bat')

    context = Context({'cmds': [cmd1, cmd2, cmd3]})
    step = AsyncCmdStep('blah', context)

    with pytest.raises(MultiError) as multi_err:
        step.run_step()

    assert 'cmdOut' not in context

    errs = multi_err.value
    assert len(errs) == 3
    err1 = errs[0]
    assert type(err1) is FileNotFoundError
    if is_posixy:
        assert err1.filename == cmd1

    err2 = errs[1]
    assert type(err2) is FileNotFoundError
    if is_posixy:
        assert err2.filename == cmd2

    err3 = errs[2]
    assert type(err3) is SubprocessError
    assert err3.returncode == 1
    assert err3.cmd == [cmd3]


def test_dsl_async_cmd_error_throws_with_save_true():
    """Process returning 1 should throw CalledProcessError."""
    cmd = get_cmd('tests/testfiles/cmds/exitwitherr.sh',
                  r'tests\testfiles\cmds\exitwitherr.bat')

    context = Context({'cmds': {'run': cmd, 'save': True}})

    step = AsyncCmdStep('blah', context)

    with pytest.raises(MultiError):
        step.run_step()

    out = context['cmdOut']
    assert len(out) == 1
    result = out[0]
    assert result.returncode == 1
    assert not result.stdout
    assert result.stderr == 'arb err here'


def test_dsl_async_cmd_error_spawning_throws_with_save_true():
    """Raise FileNotFoundError when cannot initiate the subprocess."""
    cmd = get_cmd('tests/testfiles/cmds/xxxx1',
                  r'tests\testfiles\cmds\xxxx1')

    context = Context({'cmds': {'run': [cmd], 'save': True}})

    step = AsyncCmdStep('blah', context)

    with pytest.raises(MultiError):
        step.run_step()

    out = context['cmdOut']
    assert len(out) == 1
    result = out[0]
    assert type(result) is FileNotFoundError
    if is_posixy:
        assert result.filename == cmd


def test_dsl_async_cmd_run_has_list_input_with_complex_args():
    """List input to run complex dict works with multiple args."""
    cmd1 = get_cmd('tests/testfiles/cmds/args.sh',
                   r'tests\testfiles\cmds\args.bat')

    cmd2 = get_cmd('tests/testfiles/cmds/args2.sh',
                   r'tests\testfiles\cmds\args2.bat')

    context = Context({'a': 'one',
                       'b': 'two two',
                      'c': 'three',
                       'd': cmd1,
                       'e': cmd2,
                       'cmds': {
                           'run': [
                               '{d} {a} "{b}" {c}',
                               '{e} four "five six" seven'],
                            'save': False}
                       })

    step = AsyncCmdStep('blah', context)
    step.run_step()

    assert 'cmdOut' not in context


def test_dsl_async_cmd_list_input_with_simple_cmd_strings():
    """List input with command string works with multiple args."""
    cmd1 = get_cmd('tests/testfiles/cmds/args.sh',
                   r'tests\testfiles\cmds\args.bat')

    cmd2 = get_cmd('tests/testfiles/cmds/args2.sh',
                   r'tests\testfiles\cmds\args2.bat')

    context = Context({'a': 'one',
                       'b': 'two two',
                      'c': 'three',
                       'd': cmd1,
                       'e': cmd2,
                       'cmds': [
                           '{d} {a} "{b}" {c}',
                           '{e} four "five six" seven']})

    step = AsyncCmdStep('blah', context)
    step.run_step()

    assert 'cmdOut' not in context


def test_dsl_async_cmd_list_input_with_complex_args():
    """List input mixing command string & dict works with multiple args."""
    cmd1 = get_cmd('tests/testfiles/cmds/args.sh',
                   r'tests\testfiles\cmds\args.bat')

    cmd2 = get_cmd('tests/testfiles/cmds/args2.sh',
                   r'tests\testfiles\cmds\args2.bat')

    context = Context({'a': 'one',
                       'b': 'two two',
                      'c': 'three',
                       'd': cmd1,
                       'e': cmd2,
                       'cmds': [
                           {'run': '{d} {a} "{b}" {c}',
                            'save': False},
                           '{e} four "five six" seven']})

    step = AsyncCmdStep('blah', context)
    step.run_step()

    assert 'cmdOut' not in context


def test_dsl_async_cmd_list_input_with_simple_cmd_strings_error_on_first():
    """List input with command string works runs all but raises err on one."""
    cmd1 = get_cmd('tests/testfiles/cmds/args.sh',
                   r'tests\testfiles\cmds\args.bat')

    cmd2 = get_cmd('echo blah',
                   r'tests\testfiles\cmds\echo.bat blah')

    context = Context({'a': 'WRONG VALUE',
                       'b': 'two two',
                      'c': 'three',
                       'd': cmd1,
                       'e': cmd2,
                       'cmds': [
                           '{d} {a} "{b}" {c}',
                           '{e}']})
    step = AsyncCmdStep('blah', context)

    with pytest.raises(MultiError) as err:
        step.run_step()

    assert len(err.value) == 1
    the_err = err.value[0]

    assert the_err.cmd == [cmd1, 'WRONG', 'VALUE', 'two two', 'three']

    assert 'cmdOut' not in context
    assert len(step.commands) == 2
    cmd2_results = step.commands[1].results
    assert len(cmd2_results) == 1
    assert cmd2_results[0].cmd == cmd2.split()
    assert cmd2_results[0].returncode == 0


def test_dsl_async_cmd_list_input_with_complex_args_error_on_first():
    """List input mixing command string & dict raises err on one."""
    cmd1 = get_cmd('tests/testfiles/cmds/args.sh',
                   r'tests\testfiles\cmds\args.bat')

    cmd2 = get_cmd('tests/testfiles/cmds/args2.sh',
                   r'tests\testfiles\cmds\args2.bat')

    context = Context({'a': 'WRONG',
                       'b': 'two two',
                      'c': 'three',
                       'd': cmd1,
                       'e': cmd2,
                       'cmds': [
                           {'run': '{d} {a} "{b}" {c}',
                            'save': False},
                           '{e} four "five six" seven']})

    step = AsyncCmdStep('blah', context)
    with pytest.raises(MultiError) as err:
        step.run_step()

    assert len(err.value) == 1
    the_err = err.value[0]

    assert the_err.cmd == [cmd1, 'WRONG', 'two two', 'three']

    assert 'cmdOut' not in context
    assert len(step.commands) == 2
    cmd2_results = step.commands[1].results
    assert len(cmd2_results) == 1
    assert cmd2_results[0].cmd == [cmd2, 'four', 'five six', 'seven']
    assert cmd2_results[0].returncode == 0


def test_dsl_async_cmd_dict_input_sequence_with_cwd_interpolate():
    """Cwd plus interpolation."""
    if is_windows:
        # windows path supports front slashes IF it's absolute
        cmd = cmd_path.joinpath('pwd.bat').as_posix()
    else:
        # deliberately testing relative path resolution here
        cmd = 'testfiles/cmds/pwd.sh'

    context = Context({'k1': 'tests',
                       'cmds': {'run': cmd,
                                'save': True,
                                'cwd': '{k1}'}})

    step = AsyncCmdStep('blah', context)
    step.run_step()

    out = context['cmdOut']
    assert len(out) == 1
    result = out[0]
    assert result.returncode == 0
    assert Path(result.stdout).samefile('./tests')
    assert not result.stderr
# region is_save


def test_dsl_async_cmd_list_input_with_complex_args_error_only_first_save():
    """List input mixing command str & dict, all error but 1st save output."""
    cmd1 = get_cmd('tests/testfiles/cmds/args.sh',
                   r'tests\testfiles\cmds\args.bat')

    cmd2 = get_cmd('tests/testfiles/cmds/args2.sh',
                   r'tests\testfiles\cmds\args2.bat')

    context = Context({'a': 'WRONG',
                       'b': 'two two',
                      'c': 'three',
                       'd': cmd1,
                       'e': cmd2,
                       'cmds': [
                           {'run': '{d} {a} "{b}" {c}',
                            'save': True},
                           '{e} WRONG "five six" seven']})

    step = AsyncCmdStep('blah', context)

    with pytest.raises(MultiError) as err:
        step.run_step()

    assert len(err.value) == 2
    err1 = err.value[0]

    assert err1.cmd == [cmd1, 'WRONG', 'two two', 'three']
    assert err1.returncode == 1
    assert err1.stdout == b''
    assert err1.stderr == 'assert failed'

    err2 = err.value[1]

    # save is only true for 1st one, thus no stdout/stderr
    assert err2.cmd == [cmd2, 'WRONG', 'five six', 'seven']
    assert err2.returncode == 1
    assert err2.stdout is None
    assert err2.stderr is None

    # only the cmd with save True ends up in cmdOut
    out = context['cmdOut']
    assert len(out) == 1
    out1 = out[0]
    assert out1.cmd == [cmd1, 'WRONG', 'two two', 'three']
    assert out1.returncode == 1
    assert out1.stdout == b''
    assert out1.stderr == 'assert failed'


def test_dsl_async_cmd_run_has_list_input_save():
    """List input to run complex dict save all output."""
    cmd1 = get_cmd('echo one',
                   r'tests\testfiles\cmds\echo.bat one')

    cmd2 = get_cmd('echo "two three"',
                   r'tests\testfiles\cmds\echo.bat "two three"')

    context = Context({
        'cmds': {
            'run': [cmd1, cmd2],
            'save': True}
    })

    step = AsyncCmdStep('blah', context)
    step.run_step()

    out = context['cmdOut']
    assert len(out) == 2
    out1 = out[0]
    assert out1.returncode == 0
    assert out1.stdout == 'one'
    assert not out1.stderr

    out2 = out[1]
    assert out2.returncode == 0
    assert out2.stdout == 'two three'
    assert not out2.stderr


def test_dsl_async_cmd_list_with_multiple_runs_save():
    """List input with run in expanded, everything save to results."""
    cmd1 = get_cmd('echo one',
                   r'tests\testfiles\cmds\echo.bat one')

    cmd2 = get_cmd('echo two',
                   r'tests\testfiles\cmds\echo.bat two')

    cmd3 = get_cmd('echo three',
                   r'tests\testfiles\cmds\echo.bat three')

    cmd4 = get_cmd('echo four',
                   r'tests\testfiles\cmds\echo.bat four')

    cmd5 = get_cmd('echo five',
                   r'tests\testfiles\cmds\echo.bat five')

    context = Context({
        'cmds': [
            {'run': [cmd1, cmd2],
             'save': True},
            {'run': [cmd3, cmd4],
             'save': True},
            cmd5,
        ]
    })

    step = AsyncCmdStep('blah', context)
    step.run_step()

    out = context['cmdOut']
    assert len(out) == 4

    out1 = out[0]
    assert out1.returncode == 0
    assert out1.stdout == 'one'
    assert not out1.stderr

    out2 = out[1]
    assert out2.returncode == 0
    assert out2.stdout == 'two'
    assert not out2.stderr

    out3 = out[2]
    assert out3.returncode == 0
    assert out3.stdout == 'three'
    assert not out3.stderr

    out4 = out[3]
    assert out4.returncode == 0
    assert out4.stdout == 'four'
    assert not out4.stderr


def test_dsl_async_cmd_run_has_list_input_save_dev_null():
    """List input to run complex dict save all output to /dev/null."""
    cmd1 = get_cmd('echo one',
                   r'tests\testfiles\cmds\echo.bat one')

    cmd2 = get_cmd('echo two three',
                   r'tests\testfiles\cmds\echo.bat "two three"')

    context = Context({
        'cmds': {
            'run': [cmd1, cmd2],
            'stdout': '/dev/null',
            'stderr': '/dev/null'}
    })

    step = AsyncCmdStep('blah', context)
    step.run_step()

    assert 'cmdOut' not in context

# region encoding


powershell = "powershell -ExecutionPolicy Bypass -NoLogo -NoProfile"


def test_dsl_async_cmd_run_save_with_encoding():
    """Encoding works."""
    if is_windows:
        cmd = f"{powershell} ./tests/testfiles/cmds/bytes-out.ps1"
        encoding = 'utf-16le'
    else:
        cmd = 'tests/testfiles/cmds/bytes-out.sh'
        encoding = 'utf-16'

    context = Context({
        'cmds': {
            'run': [cmd],
            'save': True,
            'encoding': encoding}
    })

    step = AsyncCmdStep('blah', context)
    step.run_step()

    out = context['cmdOut']
    assert len(out) == 1

    result = context['cmdOut'][0]
    assert result.returncode == 0
    assert result.stdout == 'one two three'
    assert result.stderr == 'four'


def test_dsl_async_cmd_run_save_with_binary_mode():
    """Binary mode does not strip lf."""
    if is_windows:
        # the literal \ escaping the " is NB
        ps_cmd = """\
$out=[System.Text.Encoding]::Unicode.GetBytes(\\"one`n\\");
[System.Console]::OpenStandardOutput().Write($out, 0, $out.Length);
$err=[System.Text.Encoding]::Unicode.GetBytes(\\"two`n\\");
[System.Console]::OpenStandardError().Write($err, 0, $err.Length);"""

        cmd = f'powershell -NoLogo -NoProfile -Command {ps_cmd}'
    else:
        cmd = 'tests/testfiles/cmds/bytes-out-lf.sh'

    context = Context({
        'cmds': {
            'run': cmd,
            'save': True,
            'bytes': True}
    })

    step = AsyncCmdStep('blah', context)
    step.run_step()

    out = context['cmdOut']
    assert len(out) == 1

    result = context['cmdOut'][0]
    assert result.returncode == 0
    assert result.stdout.decode('utf-16') == 'one\n'
    assert result.stderr.decode('utf-16') == 'two\n'
# endregion encoding

# region serial


def test_dsl_async_cmd_serial_simple_syntax_save():
    """Serial with concurrent with save."""
    echo = get_cmd('tests/testfiles/cmds/echo-out-and-err.sh',
                   r'tests\testfiles\cmds\echo-out-and-err.bat')

    context = Context({
        'cmds': {
            'run': [f'{echo} A', [f'{echo} B.1', f'{echo} B.2'], f'{echo} C'],
            'save': True}
    })

    step = AsyncCmdStep('blah', context)
    step.run_step()

    out = context['cmdOut']
    assert len(out) == 3

    outA = out[0]
    assert outA.stdout == 'stdout A'
    assert outA.stderr == 'stderr A'

    outB = out[1]
    assert len(outB) == 2
    outB1 = outB[0]
    assert outB1.stdout == 'stdout B.1'
    assert outB1.stderr == 'stderr B.1'

    outB2 = outB[1]
    assert outB2.stdout == 'stdout B.2'
    assert outB2.stderr == 'stderr B.2'

    outC = out[2]
    assert outC.stdout == 'stdout C'
    assert outC.stderr == 'stderr C'


def test_dsl_async_cmd_serial_multiple_expanded_syntax_save():
    """Serial with concurrent with save."""
    echo = get_cmd('tests/testfiles/cmds/echo-out-and-err.sh',
                   r'tests\testfiles\cmds\echo-out-and-err.bat')

    context = Context({
        'cmds': [{
            'run': [f'{echo} A', [f'{echo} B.1', f'{echo} B.2'], f'{echo} C'],
            'save': True},
            {
            'run': [[f'{echo} D.1', f'{echo} D.2']],
            'save': True}
        ]
    })

    step = AsyncCmdStep('blah', context)
    step.run_step()

    out = context['cmdOut']
    assert len(out) == 4

    outA = out[0]
    assert outA.stdout == 'stdout A'
    assert outA.stderr == 'stderr A'

    outB = out[1]
    assert len(outB) == 2
    outB1 = outB[0]
    assert outB1.stdout == 'stdout B.1'
    assert outB1.stderr == 'stderr B.1'

    outB2 = outB[1]
    assert outB2.stdout == 'stdout B.2'
    assert outB2.stderr == 'stderr B.2'

    outC = out[2]
    assert outC.stdout == 'stdout C'
    assert outC.stderr == 'stderr C'

    outD = out[3]
    assert len(outD) == 2
    outD1 = outD[0]
    assert outD1.stdout == 'stdout D.1'
    assert outD1.stderr == 'stderr D.1'

    outD2 = outD[1]
    assert outD2.stdout == 'stdout D.2'
    assert outD2.stderr == 'stderr D.2'


def test_dsl_async_cmd_serial_simple_syntax_save_with_error_spawning():
    """Serial does not continue after error spawning subprocess."""
    echo = get_cmd('tests/testfiles/cmds/echo-out-and-err.sh',
                   r'tests\testfiles\cmds\echo-out-and-err.bat')

    cmd_not_found = get_cmd('tests/testfiles/cmds/xxx1',
                            r'tests\testfiles\cmds\xxx1')
    context = Context({
        'cmds': {
            'run': [f'{echo} A',
                    [f'{echo} B.1', cmd_not_found, f'{echo} B.3'],
                    f'{echo} C'],
            'save': True}
    })

    step = AsyncCmdStep('blah', context)
    with pytest.raises(MultiError):
        step.run_step()

    out = context['cmdOut']
    assert len(out) == 3

    outA = out[0]
    assert outA.stdout == 'stdout A'
    assert outA.stderr == 'stderr A'

    outB = out[1]
    assert len(outB) == 2
    outB1 = outB[0]
    assert outB1.stdout == 'stdout B.1'
    assert outB1.stderr == 'stderr B.1'

    outB2 = outB[1]
    assert type(outB2) is FileNotFoundError
    if is_posixy:
        assert outB2.filename == cmd_not_found

    outC = out[2]
    assert outC.stdout == 'stdout C'
    assert outC.stderr == 'stderr C'


def test_dsl_async_cmd_serial_simple_syntax_save_with_error_return_1():
    """Serial does not continue after subprocess return code is 1."""
    echo = get_cmd('tests/testfiles/cmds/echo-out-and-err.sh',
                   r'tests\testfiles\cmds\echo-out-and-err.bat')

    cmd_return_1 = get_cmd('tests/testfiles/cmds/exitwitherr.sh',
                           r'tests\testfiles\cmds\exitwitherr.bat')

    context = Context({
        'cmds': {
            'run': [f'{echo} A',
                    [f'{echo} B.1', cmd_return_1, f'{echo} B.3'],
                    f'{echo} C'],
            'save': True}
    })

    step = AsyncCmdStep('blah', context)
    with pytest.raises(MultiError) as err:
        step.run_step()

    assert len(err.value) == 1
    the_err = err.value[0]
    assert the_err.cmd == [cmd_return_1]
    assert the_err.stderr == 'arb err here'

    out = context['cmdOut']
    assert len(out) == 3

    outA = out[0]
    assert outA.stdout == 'stdout A'
    assert outA.stderr == 'stderr A'

    outB = out[1]
    assert len(outB) == 2
    outB1 = outB[0]
    assert outB1.stdout == 'stdout B.1'
    assert outB1.stderr == 'stderr B.1'

    outB2 = outB[1]
    assert outB2.returncode == 1
    assert outB2.cmd == [cmd_return_1]
    assert not outB2.stdout
    assert outB2.stderr == 'arb err here'

    outC = out[2]
    assert outC.stdout == 'stdout C'
    assert outC.stderr == 'stderr C'
# endregion serial

# endregion is_save

# endregion actual subprocess spawning



"""pipelinerunner.py unit tests."""
from pathlib import Path
from unittest.mock import Mock

import pytest

from pypyr.context import Context
from pypyr.errors import (ConfigError, ContextError)
from pypyr.pipeline import Pipeline
from pypyr.pipelinerunner import run

# region fixtures


def new_pipe_and_args_wrapper(mock_pipe):
    """Return ref to Pipeline new_pipe_and_args, overriding the cls arg.

    Reason is normal patch does not patch out the cls arg on a class method.

    Intercept the factory method, passing in a mock as type to instantiate.

    Args:
        mock_pipe: Replace the 1st arg (i.e cls) to the classmethod with me.
    """
    # get the reference to the underlying method before it's patched
    og_ref = Pipeline.new_pipe_and_args.__func__

    def new_pipe_and_args(*args, **kwargs):
        # the first arg is cls - for which we're substituting the mock
        return og_ref(mock_pipe, *args, **kwargs)

    return new_pipe_and_args


@pytest.fixture
def mock_pipe(monkeypatch):
    """Intercept Pipeline.new_pipe_and_args factory method."""
    mock_pipe = Mock(spec=Pipeline)
    mock_pipe._get_parse_input = Pipeline._get_parse_input

    monkeypatch.setattr('pypyr.pipelinerunner.Pipeline.new_pipe_and_args',
                        new_pipe_and_args_wrapper(mock_pipe))

    return mock_pipe


# endregion fixtures

# region run

def test_run_all(mock_pipe):
    """Run with maximal args in."""
    out = run(pipeline_name='arb pipe',
              args_in='arb context input',
              parse_args=True,
              dict_in={'a': 'b'},
              groups=['g'],
              success_group='sg',
              failure_group='fg',
              loader='arb loader',
              py_dir='arb/dir')

    assert type(out) is Context
    assert out == Context({'a': 'b'})
    assert not out.is_in_pipeline_scope

    mock_pipe.assert_called_once_with(
        name='arb pipe',
        context_args='arb context input',
        parse_input=True,
        groups=['g'],
        success_group='sg',
        failure_group='fg',
        loader='arb loader',
        py_dir='arb/dir'
    )

    mock_pipe.return_value.run.assert_called_once_with(out)


def test_run_minimal(mock_pipe):
    """Run with minimal args in."""
    out = run('arb pipe')

    assert type(out) is Context
    assert out == {}
    assert not out.is_in_pipeline_scope

    mock_pipe.assert_called_once_with(
        name='arb pipe',
        context_args=None,
        parse_input=True,
        groups=None,
        success_group=None,
        failure_group=None,
        loader=None,
        py_dir=None
    )

    mock_pipe.return_value.run.assert_called_once_with(out)


def test_run_no_parse_when_dict_and_no_args_in(mock_pipe):
    """Default parse to false when dict_in exists and no args_in."""
    out = run('arb pipe', dict_in={'a': 'b'})

    assert type(out) is Context
    assert out == {'a': 'b'}
    assert not out.is_in_pipeline_scope

    mock_pipe.assert_called_once_with(
        name='arb pipe',
        context_args=None,
        parse_input=False,
        groups=None,
        success_group=None,
        failure_group=None,
        loader=None,
        py_dir=None
    )

    mock_pipe.return_value.run.assert_called_once_with({'a': 'b'})


def test_run_parse_when_dict_and_args_in(mock_pipe):
    """Default parse to true when dict_in exists and args_in too."""
    out = run('arb pipe', args_in=['one', 'two'], dict_in={'a': 'b'})

    assert type(out) is Context
    assert out == {'a': 'b'}
    assert not out.is_in_pipeline_scope

    mock_pipe.assert_called_once_with(
        name='arb pipe',
        context_args=['one', 'two'],
        parse_input=True,
        groups=None,
        success_group=None,
        failure_group=None,
        loader=None,
        py_dir=None
    )

    mock_pipe.return_value.run.assert_called_once_with({'a': 'b'})


def test_run_parse_when_no_dict_and_args_in(mock_pipe):
    """Default parse to true when dict_in doesn't exists and args_in does."""
    out = run('arb pipe', args_in=['one', 'two'])

    assert type(out) is Context
    assert out == {}
    assert not out.is_in_pipeline_scope

    mock_pipe.assert_called_once_with(
        name='arb pipe',
        context_args=['one', 'two'],
        parse_input=True,
        groups=None,
        success_group=None,
        failure_group=None,
        loader=None,
        py_dir=None
    )

    mock_pipe.return_value.run.assert_called_once_with({})


def test_run_parse_when_no_dict_and_no_args_in(mock_pipe):
    """Default parse to true when dict_in doesn't exists and no args_in."""
    out = run('arb pipe')

    assert type(out) is Context
    assert out == {}
    assert not out.is_in_pipeline_scope

    mock_pipe.assert_called_once_with(
        name='arb pipe',
        context_args=None,
        parse_input=True,
        groups=None,
        success_group=None,
        failure_group=None,
        loader=None,
        py_dir=None
    )

    mock_pipe.return_value.run.assert_called_once_with({})


def test_run_parse_when_parse_args_true(mock_pipe):
    """Always run parse when parse args explicitly true."""
    out = run('arb pipe', parse_args=True, dict_in={'a': 'b'})

    assert type(out) is Context
    assert out == {'a': 'b'}
    assert not out.is_in_pipeline_scope

    mock_pipe.assert_called_once_with(
        name='arb pipe',
        context_args=None,
        parse_input=True,
        groups=None,
        success_group=None,
        failure_group=None,
        loader=None,
        py_dir=None
    )

    mock_pipe.return_value.run.assert_called_once_with({'a': 'b'})


def test_no_parse_when_parse_args_false(mock_pipe):
    """Never run parse when parse args explicitly true."""
    out = run('arb pipe', args_in=['one', 'two'], parse_args=False)

    assert type(out) is Context
    assert out == {}
    assert not out.is_in_pipeline_scope

    mock_pipe.assert_called_once_with(
        name='arb pipe',
        context_args=['one', 'two'],
        parse_input=False,
        groups=None,
        success_group=None,
        failure_group=None,
        loader=None,
        py_dir=None
    )

    mock_pipe.return_value.run.assert_called_once_with({})


def test_run_raises(mock_pipe):
    """Run raises unhandled error on pipeline failure."""
    mock_pipe.return_value.run.side_effect = ContextError('arb')

    with pytest.raises(ContextError) as err_info:
        run(pipeline_name='arb pipe',
            args_in='arb context input',
            py_dir='arb/dir')

    assert str(err_info.value) == 'arb'

    mock_pipe.assert_called_once_with(
        name='arb pipe',
        context_args='arb context input',
        parse_input=True,
        groups=None,
        success_group=None,
        failure_group=None,
        loader=None,
        py_dir='arb/dir'
    )

    mock_pipe.return_value.run.assert_called_once_with({})

# endregion run

# region shortcuts


def test_run_shortcut_not_found(mock_pipe, monkeypatch):
    """Shortcuts configured but requested shortcut not found."""
    shortcuts = {'xxx': {
        'pipeline_name': 'sc pipe'
    }}
    monkeypatch.setattr('pypyr.config.config.shortcuts', shortcuts)
    out = run(pipeline_name='arb pipe',
              args_in='arb context input',
              parse_args=True,
              dict_in={'a': 'b'},
              groups=['g'],
              success_group='sg',
              failure_group='fg',
              loader='arb loader',
              py_dir='arb/dir')

    assert type(out) is Context
    assert out == Context({'a': 'b'})
    assert not out.is_in_pipeline_scope

    mock_pipe.assert_called_once_with(
        name='arb pipe',
        context_args='arb context input',
        parse_input=True,
        groups=['g'],
        success_group='sg',
        failure_group='fg',
        loader='arb loader',
        py_dir='arb/dir'
    )

    mock_pipe.return_value.run.assert_called_once_with(out)

    assert shortcuts == {'xxx': {
        'pipeline_name': 'sc pipe'
    }}


def test_run_shortcut_all(mock_pipe, monkeypatch):
    """Run shortcut with maximal shortcut options."""
    shortcuts = {'arb pipe': {
        'pipeline_name': 'sc pipe',
        'parser_args': ['sc', 'context', 'input'],
        'skip_parse': False,
        'args': {'a': 'updated', 'c': 'd'},
        'groups': ['sc g'],
        'success': 'sc sg',
        'failure': 'sc fg',
        'loader': 'sc loader',
        'py_dir': 'sc/dir'
    }}

    monkeypatch.setattr('pypyr.config.config.shortcuts', shortcuts)
    out = run(pipeline_name='arb pipe')

    assert type(out) is Context
    assert out == Context({'a': 'updated', 'c': 'd'})
    assert not out.is_in_pipeline_scope

    mock_pipe.assert_called_once_with(
        name='sc pipe',
        context_args=['sc', 'context', 'input'],
        parse_input=True,
        groups=['sc g'],
        success_group='sc sg',
        failure_group='sc fg',
        loader='sc loader',
        py_dir=Path('sc/dir')
    )

    mock_pipe.return_value.run.assert_called_once_with(out)

    # run shouldn't mutate the shared original config
    assert shortcuts == {'arb pipe': {
        'pipeline_name': 'sc pipe',
        'parser_args': ['sc', 'context', 'input'],
        'skip_parse': False,
        'args': {'a': 'updated', 'c': 'd'},
        'groups': ['sc g'],
        'success': 'sc sg',
        'failure': 'sc fg',
        'loader': 'sc loader',
        'py_dir': 'sc/dir'
    }}


def test_run_shortcut_all_ignore_func_args(mock_pipe, monkeypatch):
    """Run shortcut with maximal args superseding func args."""
    shortcuts = {'arb pipe': {
        'pipeline_name': 'sc pipe',
        'parser_args': ['sc', 'context', 'input'],
        'skip_parse': False,
        'args': {'a': 'b', 'c': 'd'},
        'groups': ['sc g'],
        'success': 'sc sg',
        'failure': 'sc fg',
        'loader': 'sc loader',
        'py_dir': 'sc/dir'
    }}

    monkeypatch.setattr('pypyr.config.config.shortcuts', shortcuts)
    out = run(pipeline_name='arb pipe',
              args_in=['arb', 'context', 'input'],
              parse_args=True,
              dict_in={'a': 'updated', 'e': 'f'},
              groups=['g'],
              success_group='sg',
              failure_group='fg',
              loader='arb loader',
              py_dir='arb/dir')

    assert type(out) is Context
    assert out == Context({'a': 'updated', 'c': 'd', 'e': 'f'})
    assert not out.is_in_pipeline_scope

    mock_pipe.assert_called_once_with(
        name='sc pipe',
        context_args=['sc', 'context', 'input', 'arb', 'context', 'input'],
        parse_input=True,
        groups=['sc g'],
        success_group='sc sg',
        failure_group='sc fg',
        loader='sc loader',
        py_dir=Path('sc/dir')
    )

    mock_pipe.return_value.run.assert_called_once_with(out)

    # run shouldn't mutate the shared original config
    assert shortcuts == {'arb pipe': {
        'pipeline_name': 'sc pipe',
        'parser_args': ['sc', 'context', 'input'],
        'skip_parse': False,
        'args': {'a': 'b', 'c': 'd'},
        'groups': ['sc g'],
        'success': 'sc sg',
        'failure': 'sc fg',
        'loader': 'sc loader',
        'py_dir': 'sc/dir'
    }}


def test_run_shortcut_minimal_fallback_func_args(mock_pipe, monkeypatch):
    """Run shortcut with not set args overwritten by func args."""
    shortcuts = {'arb pipe': {
        'pipeline_name': 'sc pipe'
    }}

    monkeypatch.setattr('pypyr.config.config.shortcuts', shortcuts)
    out = run(pipeline_name='arb pipe',
              args_in=['arb', 'context', 'input'],
              parse_args=True,
              dict_in={'a': 'b', 'e': 'f'},
              groups=['g'],
              success_group='sg',
              failure_group='fg',
              loader='arb loader',
              py_dir='arb/dir')

    assert type(out) is Context
    assert out == Context({'a': 'b', 'e': 'f'})
    assert not out.is_in_pipeline_scope

    mock_pipe.assert_called_once_with(
        name='sc pipe',
        context_args=['arb', 'context', 'input'],
        parse_input=True,
        groups=['g'],
        success_group='sg',
        failure_group='fg',
        loader='arb loader',
        py_dir='arb/dir'
    )

    mock_pipe.return_value.run.assert_called_once_with(out)

    # run shouldn't mutate the shared original config
    assert shortcuts == {'arb pipe': {
        'pipeline_name': 'sc pipe'}}


def test_run_shortcut_minimal(mock_pipe, monkeypatch):
    """Run shortcut with minimal inputs."""
    shortcuts = {'arb pipe': {
        'pipeline_name': 'sc pipe'
    }}

    monkeypatch.setattr('pypyr.config.config.shortcuts', shortcuts)
    out = run(pipeline_name='arb pipe')

    assert type(out) is Context
    assert out == {}
    assert not out.is_in_pipeline_scope

    mock_pipe.assert_called_once_with(
        name='sc pipe',
        context_args=None,
        parse_input=True,
        groups=None,
        success_group=None,
        failure_group=None,
        loader=None,
        py_dir=None
    )

    mock_pipe.return_value.run.assert_called_once_with(out)

    # run shouldn't mutate the shared original config
    assert shortcuts == {'arb pipe': {
        'pipeline_name': 'sc pipe'
    }}


def test_run_shortcut_parse_args(mock_pipe, monkeypatch):
    """Run shortcut bypasses parse_args from func input."""
    shortcuts = {'arb pipe': {
        'pipeline_name': 'sc pipe'
    }}

    monkeypatch.setattr('pypyr.config.config.shortcuts', shortcuts)
    out = run(pipeline_name='arb pipe', parse_args=False)

    assert type(out) is Context
    assert out == {}
    assert not out.is_in_pipeline_scope

    # when neither parser_args and args set, default True on parse_input.
    mock_pipe.assert_called_once_with(
        name='sc pipe',
        context_args=None,
        parse_input=True,
        groups=None,
        success_group=None,
        failure_group=None,
        loader=None,
        py_dir=None
    )

    mock_pipe.return_value.run.assert_called_once_with(out)

    # run shouldn't mutate the shared original config
    assert shortcuts == {'arb pipe': {
        'pipeline_name': 'sc pipe'
    }}


def test_run_shortcut_skip_parse(mock_pipe, monkeypatch):
    """Run shortcut with skip_parse true."""
    shortcuts = {'arb pipe': {
        'pipeline_name': 'sc pipe',
        'skip_parse': True
    }}

    monkeypatch.setattr('pypyr.config.config.shortcuts', shortcuts)
    out = run(pipeline_name='arb pipe')

    assert type(out) is Context
    assert out == {}
    assert not out.is_in_pipeline_scope

    mock_pipe.assert_called_once_with(
        name='sc pipe',
        context_args=None,
        parse_input=False,
        groups=None,
        success_group=None,
        failure_group=None,
        loader=None,
        py_dir=None
    )

    mock_pipe.return_value.run.assert_called_once_with(out)

    # run shouldn't mutate the shared original config
    assert shortcuts == {'arb pipe': {
        'pipeline_name': 'sc pipe',
        'skip_parse': True
    }}


def test_run_shortcut_groups_str(mock_pipe, monkeypatch):
    """Run shortcut with str groups converting to list."""
    shortcuts = {'arb pipe': {
        'pipeline_name': 'sc pipe',
        'groups': 'sc g'
    }}

    monkeypatch.setattr('pypyr.config.config.shortcuts', shortcuts)
    out = run(pipeline_name='arb pipe')

    assert type(out) is Context
    assert out == {}
    assert not out.is_in_pipeline_scope

    mock_pipe.assert_called_once_with(
        name='sc pipe',
        context_args=None,
        parse_input=True,
        groups=['sc g'],
        success_group=None,
        failure_group=None,
        loader=None,
        py_dir=None
    )

    mock_pipe.return_value.run.assert_called_once_with(out)

    # run shouldn't mutate the shared original config
    assert shortcuts == {'arb pipe': {
        'pipeline_name': 'sc pipe',
        'groups': 'sc g'
    }}


def test_run_shortcut_no_pipe_name(mock_pipe, monkeypatch):
    """Run shortcut raises friendly error if no pipe name."""
    shortcuts = {'arb pipe': {
        'pipeline_nameX': 'sc pipe'
    }}

    monkeypatch.setattr('pypyr.config.config.shortcuts', shortcuts)

    with pytest.raises(ConfigError) as err:
        run(pipeline_name='arb pipe')

    assert str(err.value) == (
        "shortcut 'arb pipe' has no pipeline_name set. You must set "
        "pipeline_name for this shortcut in config so that pypyr knows which "
        "pipeline to run.")

    mock_pipe.assert_not_called()

    # run shouldn't mutate the shared original config
    assert shortcuts == {'arb pipe': {
        'pipeline_nameX': 'sc pipe'
    }}


def test_run_shortcut_parser_args_str_raises(mock_pipe, monkeypatch):
    """Run shortcut raises friendly error if pipeArg is string."""
    shortcuts = {'arb pipe': {
        'pipeline_name': 'sc pipe',
        'parser_args': 'arb'
    }}

    monkeypatch.setattr('pypyr.config.config.shortcuts', shortcuts)

    with pytest.raises(ConfigError) as err:
        run(pipeline_name='arb pipe')

    assert str(err.value) == (
        "shortcut 'arb pipe' parser_args should be a list, not a string.")

    mock_pipe.assert_not_called()

    # run shouldn't mutate the shared original config
    assert shortcuts == {'arb pipe': {
        'pipeline_name': 'sc pipe',
        'parser_args': 'arb'
    }}
# endregion shortcuts



"""Unit tests for toml.py."""
import io
from unittest.mock import mock_open, patch

from pypyr.context import Context
import pypyr.toml as toml


def test_toml_file_roundtrip():
    """Read toml file, edit it, dump it."""
    in_bytes = b"""[table]
foo = "bar"  # String
baz = 13

[table2]
array = [1, 2, 3]
"""

    # read
    with patch('pypyr.toml.open',
               mock_open(read_data=in_bytes)) as mocked_open:
        payload = toml.read_file('arb/path.in')

    mocked_open.assert_called_once_with('arb/path.in', 'rb')

    assert payload['table']['baz'] == 13
    assert payload['table2']['array'] == [1, 2, 3]

    # add like you would to a normall dict
    payload['table3'] = {'a': 'b', 'c': True}

    # write
    with io.BytesIO() as out_bytes:
        with patch('pypyr.toml.open', mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_bytes.write
            toml.write_file('arb/out.toml', payload)
        out_str = out_bytes.getvalue().decode()

    mock_output.assert_called_once_with('arb/out.toml', 'wb')

    # round-tripped serialization has table3 as added from python dict obj
    assert out_str == """[table]
foo = "bar"
baz = 13

[table2]
array = [
    1,
    2,
    3,
]

[table3]
a = "b"
c = true
"""


def test_toml_context_interop():
    """Toml object works with pypyr Context."""
    in_bytes = b"""k1 = "v1"
    k2 = "start{k1}end"
    k3 = 3

    [table]
    array = [1, 2, "{k3}"]
    """

    # read
    with patch('pypyr.toml.open',
               mock_open(read_data=in_bytes)) as mocked_open:
        toml_obj = toml.read_file('arb/path.in')

    mocked_open.assert_called_once_with('arb/path.in', 'rb')

    # load toml obj into Context
    context = Context(toml_obj)

    # toml obj works with context formatting/merging methods
    assert context.get_formatted('k2') == 'startv1end'
    assert context.get_formatted_value(context['table']) == {'array':
                                                             [1, 2, 3]}

    context.merge({'table': {'array': [4, 5]}})
    assert context.get_formatted_value(context['table']['array']) == [1,
                                                                      2,
                                                                      3,
                                                                      4,
                                                                      5]



"""Context parser that returns a dict-like from a toml file."""
import logging

import pypyr.toml as toml

logger = logging.getLogger(__name__)


def get_parsed_context(args):
    """Parse input as path to a toml file, returns dict of toml contents."""
    logger.debug("starting")
    if not args:
        logger.debug("pipeline invoked without context arg set. For this toml "
                     "parser you're looking for something like: $ pypyr "
                     "pipelinename ./myfile.toml")
        return None

    path = ' '.join(args)
    logger.debug("attempting to open file: %s", path)
    payload = toml.read_file(path)

    # no special check whether top-level is mapping necessary, by spec toml
    # can only have mapping (key-value pairs or table) at top level.

    logger.debug("toml file parsed. Count: %d", len(payload))

    logger.debug("done")
    return payload



"""pypyr step gets vars from config and writes these into context."""
import logging

from pypyr.config import config

logger = logging.getLogger(__name__)


def run_step(context):
    """Write vars from config into context.

    Args:
        Context is a dictionary or dictionary-like.

    Returns: None
    """
    logger.debug("started")

    if config.vars:
        context.update(config.vars)
        logger.debug(
            f"written {len(config.vars)} variables from config.vars into "
            "context.")
    else:
        logger.debug(
            "config.vars is empty, so there is nothing to write to context.")

    logger.debug("done")



"""loadercache.py unit tests."""
from unittest.mock import call, patch, Mock

import pytest

import pypyr.cache.loadercache as loadercache
from pypyr.errors import (PipelineDefinitionError,
                          PipelineNotFoundError,
                          PyModuleNotFoundError)
from pypyr.pipedef import PipelineDefinition, PipelineInfo


# region load_the_loader


def test_load_the_loader_module_not_found():
    """Module not found raises."""
    with pytest.raises(PyModuleNotFoundError):
        loadercache.load_the_loader('blah-blah-unlikely-name')


def test_load_the_loader_attr_not_found():
    """Attribute not found raises."""
    with pytest.raises(AttributeError):
        loadercache.load_the_loader(__name__)


def test_load_the_loader_ok():
    """Module with attribute returns function."""
    loader = loadercache.load_the_loader('tests.arbpack.arbloader')
    assert isinstance(loader, loadercache.Loader)
    assert loader.name == 'tests.arbpack.arbloader'
    response = loader.get_pipeline(name='arb', parent='/arb')

    assert response == PipelineDefinition(
        pipeline={'pipeline_name': 'arb', 'parent': '/arb'},
        info=PipelineInfo('arb', 'tests.arbpack.arbloader', '/arb'))

    # 2nd time gets from cache
    response2 = loader.get_pipeline(name='arb', parent='/arb')
    assert response == response2
    assert response is response2
    assert len(loader._pipeline_cache._cache) == 1

    # name+parent is unique
    response3 = loader.get_pipeline(name='arb', parent='/arb2')
    assert response3 == PipelineDefinition(
        pipeline={'pipeline_name': 'arb', 'parent': '/arb2'},
        info=PipelineInfo('arb', 'tests.arbpack.arbloader', '/arb2'))
    assert len(loader._pipeline_cache._cache) == 2

    response4 = loader.get_pipeline(name='arb2', parent='/arb2')
    assert response4 == PipelineDefinition(
        pipeline={'pipeline_name': 'arb2', 'parent': '/arb2'},
        info=PipelineInfo('arb2', 'tests.arbpack.arbloader', '/arb2'))
    assert len(loader._pipeline_cache._cache) == 3

    response5 = loader.get_pipeline(name='arb2', parent=None)
    assert response5 == PipelineDefinition(
        pipeline={'pipeline_name': 'arb2', 'parent': None},
        info=PipelineInfo('arb2', 'tests.arbpack.arbloader', None))
    assert len(loader._pipeline_cache._cache) == 4


def test_pipeline_not_found_by_loader():
    """Pipeline not found raises."""
    with pytest.raises(PipelineNotFoundError):
        loadercache.LoaderCache().get_pype_loader().get_pipeline(
            name='blah-blah-unlikely-name',
            parent=None)

# endregion load_the_loader

# region LoaderCache

# region LoaderCache: get_pype_loader


def test_get_pype_loader_default():
    """Loader defaults to fileloader."""
    with patch('pypyr.moduleloader.get_module') as mock_get_module:
        mock_get_def = Mock()
        mock_get_def.return_value = Mock(spec=dict())
        mock_get_module.return_value.get_pipeline_definition = mock_get_def
        loader = loadercache.LoaderCache().get_pype_loader()

    mock_get_module.assert_called_once_with('pypyr.loaders.file')
    assert loader.name == 'pypyr.loaders.file'

    loader.get_pipeline('arb', 'parent')
    mock_get_def.assert_called_once_with(pipeline_name='arb', parent='parent')


def test_get_pype_loader_specified_returning_pipe_def():
    """Loader passes loader arg and works with returned PipelineDefinition."""
    with patch('pypyr.moduleloader.get_module') as mock_get_module:
        mock_get_def = Mock()
        mock_get_def.return_value = PipelineDefinition(
            pipeline={'my': 'pipe'},
            info=PipelineInfo('pipename', 'arbloader', 'parent'))

        mock_get_module.return_value.get_pipeline_definition = mock_get_def
        loader = loadercache.LoaderCache().get_pype_loader('arbloader')

    mock_get_module.assert_called_once_with('arbloader')
    assert loader.name == 'arbloader'

    pipeline = loader.get_pipeline('arb', None)
    assert pipeline == PipelineDefinition(
        pipeline={'my': 'pipe'},
        info=PipelineInfo('pipename', 'arbloader', 'parent'))
    mock_get_def.assert_called_once_with(pipeline_name='arb', parent=None)


def test_get_pype_loader_specified_wraps_in_pipedef():
    """Loader passes loader arg and wraps response in PipelineDefinition."""
    with patch('pypyr.moduleloader.get_module') as mock_get_module:
        mock_get_def = Mock()
        mock_get_def.return_value = {'a': 'b'}

        mock_get_module.return_value.get_pipeline_definition = mock_get_def
        loader = loadercache.LoaderCache().get_pype_loader('arbloader')

    mock_get_module.assert_called_once_with('arbloader')
    assert loader.name == 'arbloader'

    pipeline = loader.get_pipeline('arb', 'parent')
    assert pipeline == PipelineDefinition(
        pipeline={'a': 'b'},
        info=PipelineInfo('arb', 'arbloader', 'parent'))

    mock_get_def.assert_called_once_with(pipeline_name='arb', parent='parent')


def test_get_pype_loader_specified_raises_error_on_bad_yaml():
    """Raise error on get_pipeline where top-level yaml malformed."""
    with patch('pypyr.moduleloader.get_module') as mock_get_module:
        mock_get_def = Mock()
        # top-level yaml must be a dict, not a list.
        mock_get_def.return_value = ['a', 'b']

        mock_get_module.return_value.get_pipeline_definition = mock_get_def
        loader = loadercache.LoaderCache().get_pype_loader('arbloader')

    mock_get_module.assert_called_once_with('arbloader')
    assert loader.name == 'arbloader'

    with pytest.raises(PipelineDefinitionError):
        loader.get_pipeline('arb', 'parent')

    mock_get_def.assert_called_once_with(pipeline_name='arb', parent='parent')
# endregion LoaderCache: get_pype_loader

# region LoaderCache: clear_pipes


def test_loader_clear():
    """Clear pipeline cache in Loader."""
    with patch('pypyr.moduleloader.get_module') as mock_get_module:
        mock_get_def = Mock()
        mock_get_def.return_value = Mock(spec=dict())

        mock_get_module.return_value.get_pipeline_definition = mock_get_def
        loader = loadercache.LoaderCache().get_pype_loader('arbloader')

    mock_get_module.assert_called_once_with('arbloader')
    assert loader.name == 'arbloader'

    loader.get_pipeline('arb', None)
    loader.get_pipeline('arb2', None)

    assert mock_get_def.mock_calls == [call(pipeline_name='arb', parent=None),
                                       call(pipeline_name='arb2', parent=None)]

    assert len(loader._pipeline_cache._cache) == 2
    loader.clear()
    assert len(loader._pipeline_cache._cache) == 0


def test_loadercache_clear_pipes():
    """Clear pipeline cache in LoaderCache."""
    lc = loadercache.LoaderCache()
    with patch('pypyr.moduleloader.get_module') as mock_get_module:
        mock_get_def = Mock()
        mock_get_def.return_value = Mock(spec=dict())
        mock_get_module.return_value.get_pipeline_definition = mock_get_def

        arb_loader = lc.get_pype_loader('arbloader')
        arb_loader2 = lc.get_pype_loader('arbloader2')

    mock_get_module.assert_any_call('arbloader')
    mock_get_module.assert_any_call('arbloader2')

    arb_loader.get_pipeline('arb', None)
    arb_loader.get_pipeline('arb2', None)

    arb_loader2.get_pipeline('arb3', None)

    assert mock_get_def.mock_calls == [call(pipeline_name='arb', parent=None),
                                       call(pipeline_name='arb2', parent=None),
                                       call(pipeline_name='arb3', parent=None)]

    assert len(arb_loader._pipeline_cache._cache) == 2
    assert len(arb_loader2._pipeline_cache._cache) == 1

    lc.clear_pipes('arbloader')

    assert len(arb_loader._pipeline_cache._cache) == 0
    assert len(arb_loader2._pipeline_cache._cache) == 1

    # unknown loader name doesn't error
    arb_loader.get_pipeline('arb', None)
    lc.clear_pipes('doesntexist')
    assert len(arb_loader._pipeline_cache._cache) == 1
    assert len(arb_loader2._pipeline_cache._cache) == 1


def test_loadercache_clear_pipes_all():
    """Clear all pipeline cache in LoaderCache."""
    lc = loadercache.LoaderCache()
    with patch('pypyr.moduleloader.get_module') as mock_get_module:
        mock_get_def = Mock()
        mock_get_def.return_value = Mock(spec=dict())
        mock_get_module.return_value.get_pipeline_definition = mock_get_def
        arb_loader = lc.get_pype_loader('arbloader')
        arb_loader2 = lc.get_pype_loader('arbloader2')

    mock_get_module.mock_calls == [call('arbloader'), call('arbloader2')]

    arb_loader.get_pipeline('arb', None)
    arb_loader.get_pipeline('arb2', None)

    arb_loader2.get_pipeline('arb3', None)

    assert mock_get_def.mock_calls == [call(pipeline_name='arb', parent=None),
                                       call(pipeline_name='arb2', parent=None),
                                       call(pipeline_name='arb3', parent=None)]

    assert len(arb_loader._pipeline_cache._cache) == 2
    assert len(arb_loader2._pipeline_cache._cache) == 1

    lc.clear_pipes()

    assert len(arb_loader._pipeline_cache._cache) == 0
    assert len(arb_loader2._pipeline_cache._cache) == 0
# endregion LoaderCache: clear_pipes
# endregion LoaderCache



"""cli.py unit tests."""
from pathlib import Path
from unittest.mock import patch

import pytest

import pypyr.cli


@patch('pypyr.config.config.init')
def test_main_pass_with_sysargv_context_positional(mock_config_init):
    """Invoke from cli sets sys.argv, check assigns correctly to args."""
    arg_list = ['pypyr',
                'blah',
                'ctx string',
                '--loglevel',
                '50',
                '--dir',
                'dir here',
                '--groups',
                'group1',
                'group 2',
                'group3',
                '--success',
                'sg',
                '--failure',
                'f g']

    with patch('sys.argv', arg_list):
        with patch('pypyr.pipelinerunner.run') as mock_pipeline_run:
            with patch('pypyr.log.logger.set_root_logger') as mock_logger:
                pypyr.cli.main()

    mock_config_init.assert_called_once()
    mock_logger.assert_called_once_with(log_level=50,
                                        log_path=None)

    mock_pipeline_run.assert_called_once_with(
        pipeline_name='blah',
        args_in=['ctx string'],
        parse_args=True,
        py_dir='dir here',
        groups=['group1', 'group 2', 'group3'],
        success_group='sg',
        failure_group='f g'
    )


@patch('pypyr.config.config.init')
def test_main_pass_with_sysargv_single_group(mock_config_init):
    """Invoke from cli sets sys.argv, check assigns correctly to group."""
    arg_list = ['pypyr',
                'blah',
                'ctx string',
                '--loglevel',
                '50',
                '--dir',
                'dir here',
                '--groups',
                'group1',
                '--success',
                'sg',
                '--failure',
                'f g']

    with patch('sys.argv', arg_list):
        with patch('pypyr.pipelinerunner.run') as mock_pipeline_run:
            with patch('pypyr.log.logger.set_root_logger') as mock_logger:
                pypyr.cli.main()

    mock_logger.assert_called_once_with(log_level=50,
                                        log_path=None)

    mock_config_init.assert_called_once()

    mock_pipeline_run.assert_called_once_with(
        pipeline_name='blah',
        args_in=['ctx string'],
        parse_args=True,
        py_dir='dir here',
        groups=['group1'],
        success_group='sg',
        failure_group='f g'
    )


@patch('pypyr.config.config.init')
def test_main_pass_with_sysargv_context_multiple_positional(mock_config_init):
    """Multiple positional arguments."""
    arg_list = ['pypyr',
                'blah',
                'ctx string',
                'arg 2',
                'arg 3',
                'arg4=hello',
                '--loglevel',
                '50',
                '--dir',
                'dir here',
                '--groups',
                'group1',
                'group 2',
                'group3',
                '--success',
                'sg',
                '--failure',
                'f g']

    with patch('sys.argv', arg_list):
        with patch('pypyr.pipelinerunner.run') as mock_pipeline_run:
            with patch('pypyr.log.logger.set_root_logger') as mock_logger:
                pypyr.cli.main()

    mock_logger.assert_called_once_with(log_level=50,
                                        log_path=None)

    mock_config_init.assert_called_once()

    mock_pipeline_run.assert_called_once_with(
        pipeline_name='blah',
        args_in=['ctx string', 'arg 2', 'arg 3', 'arg4=hello'],
        parse_args=True,
        py_dir='dir here',
        groups=['group1', 'group 2', 'group3'],
        success_group='sg',
        failure_group='f g'
    )


@patch('pypyr.config.config.init')
def test_main_pass_with_sysargv_context_positional_log_alias(mock_config_init):
    """Invoke from cli sets sys.argv with log alias."""
    arg_list = ['pypyr',
                'blah',
                'ctx string',
                '--log',
                '50',
                '--dir',
                'dir here']

    with patch('sys.argv', arg_list):
        with patch('pypyr.pipelinerunner.run') as mock_pipeline_run:
            with patch('pypyr.log.logger.set_root_logger') as mock_logger:
                pypyr.cli.main()

    mock_logger.assert_called_once_with(log_level=50,
                                        log_path=None)

    mock_config_init.assert_called_once()

    mock_pipeline_run.assert_called_once_with(
        pipeline_name='blah',
        args_in=['ctx string'],
        parse_args=True,
        py_dir='dir here',
        groups=None,
        success_group=None,
        failure_group=None
    )


@patch('pypyr.config.config.init')
def test_main_pass_with_sysargv_context_positional_abbreviations(
        mock_config_init):
    """Invoke from cli sets sys.argv with log abbreviations."""
    arg_list = ['pypyr',
                'blah',
                'ctx string',
                '--logl',
                '50',
                '--dir',
                'dir here',
                '--logp',
                '/blah']

    with patch('sys.argv', arg_list):
        with patch('pypyr.pipelinerunner.run') as mock_pipeline_run:
            with patch('pypyr.log.logger.set_root_logger') as mock_logger:
                pypyr.cli.main()

    mock_logger.assert_called_once_with(log_level=50,
                                        log_path='/blah')

    mock_config_init.assert_called_once()

    mock_pipeline_run.assert_called_once_with(
        pipeline_name='blah',
        args_in=['ctx string'],
        parse_args=True,
        py_dir='dir here',
        groups=None,
        success_group=None,
        failure_group=None
    )


@patch('pypyr.config.config.init')
def test_main_pass_with_sysargv_context_positional_flags_last(
        mock_config_init):
    """Check assigns correctly to args when positional last not first."""
    arg_list = ['pypyr',
                '--loglevel',
                '50',
                '--dir',
                'dir here',
                'blah',
                'ctx string', ]

    with patch('sys.argv', arg_list):
        with patch('pypyr.pipelinerunner.run') as mock_pipeline_run:
            with patch('pypyr.log.logger.set_root_logger') as mock_logger:
                pypyr.cli.main()

    mock_logger.assert_called_once_with(log_level=50,
                                        log_path=None)

    mock_config_init.assert_called_once()

    mock_pipeline_run.assert_called_once_with(
        pipeline_name='blah',
        args_in=['ctx string'],
        parse_args=True,
        py_dir='dir here',
        groups=None,
        success_group=None,
        failure_group=None
    )


@patch('pypyr.config.config.init')
def test_main_pass_with_sysargv_context_multiple_positional_flags_last(
        mock_config_init):
    """Check assigns correctly to multiple args when positional last."""
    arg_list = ['pypyr',
                '--loglevel',
                '50',
                '--dir',
                'dir here',
                'blah',
                'ctx string',
                'arb 2',
                'arb3=arbvalue']

    with patch('sys.argv', arg_list):
        with patch('pypyr.pipelinerunner.run') as mock_pipeline_run:
            with patch('pypyr.log.logger.set_root_logger') as mock_logger:
                pypyr.cli.main()

    mock_logger.assert_called_once_with(log_level=50,
                                        log_path=None)

    mock_config_init.assert_called_once()

    mock_pipeline_run.assert_called_once_with(
        pipeline_name='blah',
        args_in=['ctx string', 'arb 2', 'arb3=arbvalue'],
        parse_args=True,
        py_dir='dir here',
        groups=None,
        success_group=None,
        failure_group=None
    )


@patch('pypyr.config.config.init')
def test_main_pass_with_defaults_context_positional(mock_config_init):
    """Default values assigned - log 25 and cwd."""
    arg_list = ['blah',
                'ctx string']

    with patch('pypyr.pipelinerunner.run') as mock_pipeline_run:
        with patch('pypyr.log.logger.set_root_logger') as mock_logger:
            pypyr.cli.main(arg_list)

    mock_logger.assert_called_once_with(log_level=None,
                                        log_path=None)

    mock_config_init.assert_called_once()

    mock_pipeline_run.assert_called_once_with(
        pipeline_name='blah',
        args_in=['ctx string'],
        parse_args=True,
        py_dir=Path.cwd(),
        groups=None,
        success_group=None,
        failure_group=None
    )


@patch('pypyr.config.config.init')
def test_main_pass_with_no_context(mock_config_init):
    """No context is None."""
    arg_list = ['blah']

    with patch('pypyr.pipelinerunner.run') as mock_pipeline_run:
        with patch('pypyr.log.logger.set_root_logger') as mock_logger:
            pypyr.cli.main(arg_list)

    mock_logger.assert_called_once_with(log_level=None,
                                        log_path=None)

    mock_pipeline_run.assert_called_once_with(
        pipeline_name='blah',
        args_in=[],
        parse_args=True,
        py_dir=Path.cwd(),
        groups=None,
        success_group=None,
        failure_group=None
    )


@patch('pypyr.config.config.init')
def test_main_pass_with_no_context_other_flags_set(mock_config_init):
    """No context is None and other flag still work."""
    arg_list = ['blah',
                '--loglevel',
                '11']

    with patch('pypyr.pipelinerunner.run') as mock_pipeline_run:
        with patch('pypyr.log.logger.set_root_logger') as mock_logger:
            pypyr.cli.main(arg_list)

    mock_logger.assert_called_once_with(log_level=11,
                                        log_path=None)

    mock_config_init.assert_called_once()

    mock_pipeline_run.assert_called_once_with(
        pipeline_name='blah',
        args_in=[],
        parse_args=True,
        py_dir=Path.cwd(),
        groups=None,
        success_group=None,
        failure_group=None
    )


def test_pipeline_name_required():
    """Error expected if no pipeline name."""
    arg_list = ['--dir',
                'blah']

    with patch('pypyr.pipelinerunner.run'):
        with pytest.raises(SystemExit) as exit_err:
            pypyr.cli.main(arg_list)

        assert exit_err.value.code == 2


@patch('pypyr.config.config.init')
def test_interrupt_signal(mock_config_init):
    """Interrupt signal handled."""
    arg_list = ['blah',
                'ctx string']

    with patch('pypyr.pipelinerunner.run') as mock_pipeline_run:
        mock_pipeline_run.side_effect = KeyboardInterrupt()
        val = pypyr.cli.main(arg_list)
        assert val == 130

    mock_config_init.assert_called_once()


@patch('pypyr.config.config.init')
def test_arb_error(mock_config_init):
    """Arbitrary error should return 255."""
    arg_list = ['blah',
                'ctx string']

    with patch('pypyr.pipelinerunner.run') as mock_pipeline_run:
        mock_pipeline_run.side_effect = AssertionError('Test Error Mock')
        val = pypyr.cli.main(arg_list)
        assert val == 255

    mock_config_init.assert_called_once()


@patch('pypyr.config.config.init')
def test_trace_log_level_less_10(mock_config_init):
    """Log Level < 10 produces traceback on error."""
    arg_list = ['blah',
                'ctx string',
                '--loglevel',
                '5']

    with patch('pypyr.pipelinerunner.run') as mock_pipeline_run:
        with patch('traceback.print_exc') as mock_traceback:
            mock_pipeline_run.side_effect = AssertionError('Test Error Mock')
            val = pypyr.cli.main(arg_list)
            assert val == 255

    mock_traceback.assert_called_once()

    mock_config_init.assert_called_once()


@patch('pypyr.config.config.init')
def test_trace_log_level_over_10(mock_config_init):
    """Log Level > 10 doesn't produce traceback on error."""
    arg_list = ['blah',
                'ctx string',
                '--loglevel',
                '11']

    with patch('pypyr.pipelinerunner.run') as mock_pipeline_run:
        with patch('traceback.print_exc') as mock_traceback:
            mock_pipeline_run.side_effect = AssertionError('Test Error Mock')
            val = pypyr.cli.main(arg_list)
            assert val == 255

    mock_traceback.assert_not_called()
    mock_config_init.assert_called_once()


@patch('pypyr.config.config.init')
def test_trace_log_level_none(mock_config_init):
    """Log Level None doesn't produce traceback on error."""
    arg_list = ['blah',
                'ctx string']

    with patch('pypyr.pipelinerunner.run') as mock_pipeline_run:
        with patch('traceback.print_exc') as mock_traceback:
            mock_pipeline_run.side_effect = AssertionError('Test Error Mock')
            val = pypyr.cli.main(arg_list)
            assert val == 255

    mock_traceback.assert_not_called()
    mock_config_init.assert_called_once()


@patch('pypyr.config.config.init')
def test_main_pass_with_logpath(mock_config_init):
    """The logpath set to tempfile."""
    arg_list = ['blah',
                '--logpath',
                'tmp.log']

    with patch('pypyr.pipelinerunner.run') as mock_pipeline_run:
        with patch('pypyr.log.logger.set_root_logger') as mock_logger:
            pypyr.cli.main(arg_list)

    mock_logger.assert_called_once_with(log_level=None,
                                        log_path='tmp.log',)

    mock_config_init.assert_called_once()

    mock_pipeline_run.assert_called_once_with(
        pipeline_name='blah',
        args_in=[],
        parse_args=True,
        py_dir=Path.cwd(),
        groups=None,
        success_group=None,
        failure_group=None
    )



"""glob.py unit tests."""
import logging
import pytest
from unittest.mock import patch
from pypyr.context import Context
from pypyr.errors import KeyInContextHasNoValueError, KeyNotInContextError
import pypyr.steps.glob as glob_step
from tests.common.utils import patch_logger


def test_glob_no_input_raises():
    """No input context raises."""
    context = Context({
        'k1': 'v1'})

    with pytest.raises(KeyNotInContextError) as err_info:
        glob_step.run_step(context)

    assert str(err_info.value) == ("context['glob'] "
                                   "doesn't exist. It must exist for "
                                   "pypyr.steps.glob.")


def test_glob_empty_path_raises():
    """Empty input raises."""
    context = Context({
        'glob': None})

    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        glob_step.run_step(context)

    assert str(err_info.value) == ("context['glob'] must have a "
                                   "value for pypyr.steps.glob.")


def test_glob_empty_string_path_raises():
    """Empty string glob raises."""
    context = Context({
        'glob': ''})

    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        glob_step.run_step(context)

    assert str(err_info.value) == ("The glob path is an empty str")


def test_glob_empty_path_list_raises():
    """Empty glob list raises."""
    context = Context({
        'glob': []})

    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        glob_step.run_step(context)

    assert str(err_info.value) == ("The glob list has an empty str")


def test_glob_list_with_empty_raises():
    """Glob list with missing item raises."""
    context = Context({
        'glob': ['a', '', 'c']})

    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        glob_step.run_step(context)

    assert str(err_info.value) == ("The glob list has an empty str")


@patch('pypyr.utils.filesystem.get_glob')
def test_glob_single(mock_glob):
    """Single path ok."""
    context = Context({
        'ok1': 'ov1',
        'glob': './arb/x'})

    mock_glob.return_value = ['./foundfile']

    with patch_logger('pypyr.steps.glob', logging.INFO) as mock_logger_info:
        glob_step.run_step(context)

    mock_logger_info.assert_called_once_with(
        'glob checked 1 globs and saved 1 paths to globOut')
    assert context, "context shouldn't be None"
    assert len(context) == 3, "context should have 3 items"
    assert context['ok1'] == 'ov1'
    assert context['glob'] == './arb/x'
    assert context["globOut"] == ['./foundfile']

    mock_glob.assert_called_once_with('./arb/x')


@patch('pypyr.utils.filesystem.get_glob')
def test_glob_single_not_found(mock_glob):
    """Single path ok on not found."""
    context = Context({
        'ok1': 'ov1',
        'glob': './arb/x'})

    mock_glob.return_value = []

    with patch_logger(
            'pypyr.steps.glob', logging.INFO) as mock_logger_info:
        glob_step.run_step(context)

    mock_logger_info.assert_called_once_with(
        'glob checked 1 globs and saved 0 paths to globOut')

    assert context, "context shouldn't be None"
    assert len(context) == 3, "context should have 3 items"
    assert context['ok1'] == 'ov1'
    assert context['glob'] == './arb/x'
    assert context["globOut"] == []

    mock_glob.assert_called_once_with('./arb/x')


@patch('pypyr.utils.filesystem.get_glob')
def test_glob_single_with_formatting(mock_glob):
    """Single path ok with string formatting."""
    context = Context({
        'ok1': 'ov1',
        'glob': './{ok1}/x'})

    mock_glob.return_value = ['./foundfile']

    glob_step.run_step(context)

    assert context, "context shouldn't be None"
    assert len(context) == 3, "context should have 3 items"
    assert context['ok1'] == 'ov1'
    assert context['glob'] == './{ok1}/x'
    assert context["globOut"] == ['./foundfile']

    mock_glob.assert_called_once_with('./ov1/x')


@patch('pypyr.utils.filesystem.get_glob', autospec=True)
def test_glob_list(mock_glob):
    """Multiple paths ok."""
    context = Context({
        'ok1': 'ov1',
        'glob': ['./arb/x', './arb/y', './arb/z']})

    mock_glob.return_value = [
        './f1.1',
        './f2.1',
        './f2.2',
        './f2.3',
    ]

    with patch_logger('pypyr.steps.glob', logging.INFO) as mock_logger_info:
        glob_step.run_step(context)

    mock_logger_info.assert_called_once_with(
        'glob checked 3 globs and saved 4 paths to globOut')

    assert context, "context shouldn't be None"
    assert len(context) == 3, "context should have 3 items"
    assert context['ok1'] == 'ov1'
    assert context['glob'] == ['./arb/x', './arb/y', './arb/z']
    assert context["globOut"] == [
        './f1.1',
        './f2.1',
        './f2.2',
        './f2.3',
    ]

    mock_glob.assert_called_once_with(
        ['./arb/x', './arb/y', './arb/z'])


@patch('pypyr.utils.filesystem.get_glob')
def test_glob_list_none(mock_glob):
    """Multiple paths ok with none returning match."""
    context = Context({
        'ok1': 'ov1',
        'glob': ['./{ok1}/x', './arb/{ok1}', '{ok1}/arb/z']})

    mock_glob.return_value = []

    with patch_logger('pypyr.steps.glob', logging.INFO) as mock_logger_info:
        glob_step.run_step(context)

    mock_logger_info.assert_called_once_with(
        'glob checked 3 globs and saved 0 paths to globOut')

    assert context, "context shouldn't be None"
    assert len(context) == 3, "context should have 3 items"
    assert context['ok1'] == 'ov1'
    assert context['glob'] == ['./{ok1}/x', './arb/{ok1}', '{ok1}/arb/z']
    assert context["globOut"] == []

    mock_glob.assert_called_once_with(['./ov1/x',
                                       './arb/ov1',
                                       'ov1/arb/z'])



"""Platform specific handling. For example, O/S specific directories.

Although it is not a direct copy and it's been substantially adjusted to fit
with pypyr's opinionated implementation, this module still owes a LOT of
inspiration to the PlatformDirs library, licensed under the MIT license.

https://github.com/platformdirs/platformdirs

This is the PlatformDirs license notice: This is the MIT license

Copyright (c) 2010 ActiveState Software Inc.

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""  # noqa: E501
# the noqa is for line length.
# can remove __future__ once py 3.10 the lowest supported version
from __future__ import annotations
from abc import ABC, abstractmethod
import os
from pathlib import Path
import re
import sys
from typing import Type, NamedTuple
# NB: This module pypyr.platform should NOT import anything from anywhere in
# pypyr, to prevent circular imports. Reason is pypyr.config imports this, and
# pypyr.config is used all over as a fundamental ground-level dependency.


class PlatformPaths(NamedTuple):
    """Platform specific paths for config files and data directories.

    The reason for declaring the NamedTuple like this is
    a) to make type information available
    b) allow the custom __str__ override.
    """

    config_user: Path
    config_common: list[Path]
    data_dir_user: Path
    data_dir_common: list[Path]

    def __str__(self) -> str:
        """Represent as friendly string with 0 offset."""
        return self.to_str()

    def to_str(self, offset: int = 0) -> str:
        """Represent class as human friendly string.

        Args:
            offset: Pad on left with this many spaces.

        Return:
            String representation of object.
        """
        offset_str = ' ' * offset
        return (f'{offset_str}config_user: {self.config_user}\n'
                + f'{offset_str}config_common:\n'
                + ''.join(f'{offset_str}  - {p}\n' for p in self.config_common)
                + f'{offset_str}data_dir_user: {self.data_dir_user}\n'
                + f'{offset_str}data_dir_common:\n'
                + ''.join(
                    f'{offset_str}  - {p}\n' for p in self.data_dir_common)
                )


class BaseDirFinder(ABC):
    """All platform specific directory finders derive from this."""

    def __init__(self, app_name: str, config_file_name: str) -> None:
        """Initialize app_name and config_file_name."""
        self.app_name = app_name  # pypyr
        self.config_file_name = config_file_name  # config.yaml

    def get_platform_paths(self) -> PlatformPaths:
        """Construct a PlatformPaths instance specific to this platform."""
        return PlatformPaths(config_user=self.get_config_user(),
                             config_common=self.get_config_common(),
                             data_dir_user=self.get_data_user(),
                             data_dir_common=self.get_data_common())

    @abstractmethod
    def get_config_user(self) -> Path:
        """Get path to user config file."""
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def get_config_common(self) -> list[Path]:
        """Get paths to shared system-level config files."""
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def get_data_user(self) -> Path:
        """Get user data directory for this platform."""
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def get_data_common(self) -> list[Path]:
        """Get shared system-level data directory for this platform."""
        raise NotImplementedError()  # pragma: no cover


class Xdg(BaseDirFinder):
    """Find directories based on the XDG Base Dir specification."""

    def __init__(self, app_name: str, config_file_name: str):
        """Initialize defaults."""
        super().__init__(app_name, config_file_name)

        self.common_config_base_dir_default = '/etc/xdg'
        self.common_data_base_dir_default = (
            f'/usr/local/share{os.pathsep}/usr/share')

    def get_config_user(self) -> Path:
        """Get $XDG_CONFIG_HOME/config_file_name.

        Default dir is ~/.config/app_name/config_file_name.
        """
        path = os.getenv('XDG_CONFIG_HOME', '')
        if not path.strip():
            path = os.path.expanduser('~/.config')

        return self.get_pypyr_config_file_appended(path)

    def get_config_common(self) -> list[Path]:
        """Get list of dirs at $XDG_CONFIG_DIRS.

        app_name/config_file_name appended to each dir.
        """
        path = os.getenv('XDG_CONFIG_DIRS', '')
        if not path.strip():
            path = self.common_config_base_dir_default

        return [self.get_pypyr_config_file_appended(p)
                for p in path.split(os.pathsep) if p.strip()]

    def get_data_user(self) -> Path:
        """Get $XDG_DATA_HOME. Default is ~/.local/share/app_name."""
        path = os.getenv('XDG_DATA_HOME', '')
        if not path.strip():
            path = os.path.expanduser('~/.local/share')

        return Path(path, self.app_name)

    def get_data_common(self) -> list[Path]:
        """Get $XDG_DATA_DIRS.

        Default is
            /usr/local/share/app_name
            /usr/share/app_name
        """
        path = os.getenv('XDG_DATA_DIRS', '')
        if not path.strip():
            path = self.common_data_base_dir_default

        app_name = self.app_name
        return [Path(p, app_name) for p in path.split(os.pathsep) if p.strip()]

    def get_pypyr_config_file_appended(self, base_path: str) -> Path:
        """Append app_name and config_file_name to base_path."""
        return Path(base_path, self.app_name, self.config_file_name)


class MacOs(Xdg):
    """Find platform directories for MacOs.

    Largely follows XDG spec, except for the common dirs.
    """

    def __init__(self, app_name: str, config_file_name: str):
        """Initialize MacOs platform directories."""
        super().__init__(app_name, config_file_name)
        self.common_config_base_dir_default = '/Library/Application Support'
        self.common_data_base_dir_default = '/Library/Application Support'


class Windows(Xdg):
    """Find platform directories for Windows.

    Largely follows XDG spec, except for the common (i.e all user) dirs.
    """

    def __init__(self, app_name: str, config_file_name: str):
        """Initialize Windows directory finder."""
        super().__init__(app_name, config_file_name)
        common_appdata = os.getenv('ALLUSERSPROFILE', 'C:/ProgramData')

        self.common_config_base_dir_default = common_appdata
        self.common_data_base_dir_default = common_appdata


class Android(BaseDirFinder):
    """Find platform directories for Android."""

    def __init__(self, app_name, config_file_name):
        """Initialize Android directory finder."""
        super().__init__(app_name, config_file_name)
        android_dir = Path(Android._get_android_dir())
        self._config_path = android_dir.joinpath('shared_prefs',
                                                 self.app_name,
                                                 self.config_file_name)

        self._data_dir = android_dir.joinpath('files', app_name)

    def get_config_user(self) -> Path:
        """Return config directory for the user.

        e.g. ``/data/user/<userid>/<packagename>/shared_prefs/<AppName>``
        """
        return self._config_path

    def get_config_common(self) -> list[Path]:
        """Get common config path. Same as get_config_user."""
        return [self._config_path]

    def get_data_user(self) -> Path:
        """Return data directory for the user.

        e.g. ``/data/user/<userid>/<packagename>/files/<AppName>``
        """
        return self._data_dir

    def get_data_common(self) -> list[Path]:
        """Get common data dir. Same as get_data_user."""
        return [self._data_dir]

    @staticmethod
    def _get_android_dir() -> str:
        """Return the base folder for the Android OS.

        This method is wholly borrowed from PlatformDirs, with thanks!
        """
        try:
            # First try to get path to android app via pyjnius
            from jnius import autoclass  # type: ignore # noqa: SC200

            Context = autoclass('android.content.Context')  # noqa: SC200
            s: str = Context.getFilesDir().getParentFile().getAbsolutePath()
        except Exception:
            # if fails find an android folder looking path on the sys.path
            pattern = re.compile(r"/data/(data|user/\d+)/(.+)/files")
            for path in sys.path:
                if pattern.match(path):
                    s = path.split('/files')[0]
                    break
            else:
                raise OSError("Cannot find path to android app folder")
        return s


def get_platform_paths(app_name: str, config_file_name: str) -> PlatformPaths:
    """Calculate platform specific paths for this O/S.

    Args:
        app_name: Name of application. This forms part of the directory name.
        config_file_name: The name of the config file to append to config dirs.

    Returns:
        An instance of the PlatformPaths tuple, containing the paths relevant
        for this O/S.
    """
    platform_type = get_platform_dir_finder()
    dir_finder = platform_type(app_name, config_file_name)

    return dir_finder.get_platform_paths()


def get_platform_dir_finder() -> Type[BaseDirFinder]:
    """Return the BaseDirFinder specific to current O/S platform.

    The logic is pretty much to treat everything that is neither Windows, MacOs
    nor Android as POSIX. Android is checked via ANDROID_* env variables.

    This is a factory function. If you're trying to find platform specific
    paths, you should probably use get_platform_paths instead.

    Returns:
        Derived instance of BaseDirFinder appropriate to the current O/S.
    """
    current_platform = sys.platform
    platform_type: Type[BaseDirFinder]
    if (os.getenv('ANDROID_DATA') == '/data'
            and os.getenv('ANDROID_ROOT') == '/system'):
        platform_type = Android
    elif current_platform == 'win32':
        platform_type = Windows
    elif current_platform == 'darwin':
        platform_type = MacOs
    else:
        platform_type = Xdg

    return platform_type
